# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

from ctypes import CDLL, c_uint32, c_char_p, c_size_t, c_int, c_uint

from .._ffi import FFIEngineError


__all__ = [
    'CommonCrypto',
]


common_crypto_path = '/usr/lib/system/libcommonCrypto.dylib'

CommonCrypto = CDLL(common_crypto_path, use_errno=True)

try:
    CommonCrypto.CCKeyDerivationPBKDF.argtypes = [
        c_uint32,
        c_char_p,
        c_size_t,
        c_char_p,
        c_size_t,
        c_uint32,
        c_uint,
        c_char_p,
        c_size_t
    ]
    CommonCrypto.CCKeyDerivationPBKDF.restype = c_int
except (AttributeError):
    raise FFIEngineError('Error initializing ctypes')



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

from .._ffi import register_ffi
from .._types import str_cls
from ..errors import LibraryNotFoundError

import cffi


__all__ = [
    'get_error',
    'kernel32',
]


ffi = cffi.FFI()
if cffi.__version_info__ >= (0, 9):
    ffi.set_unicode(True)
ffi.cdef("""
    typedef long long LARGE_INTEGER;
    BOOL QueryPerformanceCounter(LARGE_INTEGER *lpPerformanceCount);

    typedef struct _FILETIME {
        DWORD dwLowDateTime;
        DWORD dwHighDateTime;
    } FILETIME;

    void GetSystemTimeAsFileTime(FILETIME *lpSystemTimeAsFileTime);
""")


try:
    kernel32 = ffi.dlopen('kernel32.dll')
    register_ffi(kernel32, ffi)

except (OSError) as e:
    if str_cls(e).find('cannot load library') != -1:
        raise LibraryNotFoundError('kernel32.dll could not be found')
    raise


def get_error():
    return ffi.getwinerror()



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import sys

import ctypes
from ctypes import windll, wintypes, POINTER, c_void_p, c_uint, Structure
from ctypes.wintypes import DWORD, ULONG

from .._ffi import FFIEngineError
from .._types import str_cls
from ..errors import LibraryNotFoundError


__all__ = [
    'get_error',
    'secur32',
]


try:
    secur32 = windll.secur32
except (OSError) as e:
    if str_cls(e).find('The specified module could not be found') != -1:
        raise LibraryNotFoundError('secur32.dll could not be found')
    raise

HCERTSTORE = wintypes.HANDLE
ALG_ID = c_uint
if sys.maxsize > 2 ** 32:
    ULONG_PTR = ctypes.c_uint64
else:
    ULONG_PTR = ctypes.c_ulong
SEC_GET_KEY_FN = c_void_p
LUID = c_void_p
SECURITY_STATUS = ctypes.c_ulong
SEC_WCHAR = wintypes.WCHAR

try:
    class SecHandle(Structure):
        _fields_ = [
            ('dwLower', ULONG_PTR),
            ('dwUpper', ULONG_PTR),
        ]

    CredHandle = SecHandle
    CtxtHandle = SecHandle

    class SCHANNEL_CRED(Structure):  # noqa
        _fields_ = [
            ('dwVersion', DWORD),
            ('cCreds', DWORD),
            ('paCred', c_void_p),
            ('hRootStore', HCERTSTORE),
            ('cMappers', DWORD),
            ('aphMappers', POINTER(c_void_p)),
            ('cSupportedAlgs', DWORD),
            ('palgSupportedAlgs', POINTER(ALG_ID)),
            ('grbitEnabledProtocols', DWORD),
            ('dwMinimumCipherStrength', DWORD),
            ('dwMaximumCipherStrength', DWORD),
            ('dwSessionLifespan', DWORD),
            ('dwFlags', DWORD),
            ('dwCredFormat', DWORD),
        ]

    class TimeStamp(Structure):
        _fields_ = [
            ('dwLowDateTime', DWORD),
            ('dwHighDateTime', DWORD),
        ]

    class SecBuffer(Structure):
        _fields_ = [
            ('cbBuffer', ULONG),
            ('BufferType', ULONG),
            ('pvBuffer', POINTER(ctypes.c_byte)),
        ]

    PSecBuffer = POINTER(SecBuffer)

    class SecBufferDesc(Structure):
        _fields_ = [
            ('ulVersion', ULONG),
            ('cBuffers', ULONG),
            ('pBuffers', PSecBuffer),
        ]

    class SecPkgContext_StreamSizes(Structure):  # noqa
        _fields_ = [
            ('cbHeader', ULONG),
            ('cbTrailer', ULONG),
            ('cbMaximumMessage', ULONG),
            ('cBuffers', ULONG),
            ('cbBlockSize', ULONG),
        ]

    class SecPkgContext_ConnectionInfo(Structure):  # noqa
        _fields_ = [
            ('dwProtocol', DWORD),
            ('aiCipher', ALG_ID),
            ('dwCipherStrength', DWORD),
            ('aiHash', ALG_ID),
            ('dwHashStrength', DWORD),
            ('aiExch', ALG_ID),
            ('dwExchStrength', DWORD),
        ]

    secur32.AcquireCredentialsHandleW.argtypes = [
        POINTER(SEC_WCHAR),
        POINTER(SEC_WCHAR),
        ULONG,
        POINTER(LUID),
        c_void_p,
        SEC_GET_KEY_FN,
        c_void_p,
        POINTER(CredHandle),
        POINTER(TimeStamp)
    ]
    secur32.AcquireCredentialsHandleW.restype = SECURITY_STATUS

    secur32.FreeCredentialsHandle.argtypes = [
        POINTER(CredHandle)
    ]
    secur32.FreeCredentialsHandle.restype = SECURITY_STATUS

    secur32.InitializeSecurityContextW.argtypes = [
        POINTER(CredHandle),
        POINTER(CtxtHandle),
        POINTER(SEC_WCHAR),
        ULONG,
        ULONG,
        ULONG,
        POINTER(SecBufferDesc),
        ULONG,
        POINTER(CtxtHandle),
        POINTER(SecBufferDesc),
        POINTER(ULONG),
        POINTER(TimeStamp)
    ]
    secur32.InitializeSecurityContextW.restype = SECURITY_STATUS

    secur32.FreeContextBuffer.argtypes = [
        c_void_p
    ]
    secur32.FreeContextBuffer.restype = SECURITY_STATUS

    secur32.ApplyControlToken.argtypes = [
        POINTER(CtxtHandle),
        POINTER(SecBufferDesc)
    ]
    secur32.ApplyControlToken.restype = SECURITY_STATUS

    secur32.DeleteSecurityContext.argtypes = [
        POINTER(CtxtHandle)
    ]
    secur32.DeleteSecurityContext.restype = SECURITY_STATUS

    secur32.QueryContextAttributesW.argtypes = [
        POINTER(CtxtHandle),
        ULONG,
        c_void_p
    ]
    secur32.QueryContextAttributesW.restype = SECURITY_STATUS

    secur32.EncryptMessage.argtypes = [
        POINTER(CtxtHandle),
        ULONG,
        POINTER(SecBufferDesc),
        ULONG
    ]
    secur32.EncryptMessage.restype = SECURITY_STATUS

    secur32.DecryptMessage.argtypes = [
        POINTER(CtxtHandle),
        POINTER(SecBufferDesc),
        ULONG,
        POINTER(ULONG)
    ]
    secur32.DecryptMessage.restype = SECURITY_STATUS

except (AttributeError):
    raise FFIEngineError('Error initializing ctypes')


setattr(secur32, 'ALG_ID', ALG_ID)
setattr(secur32, 'CredHandle', CredHandle)
setattr(secur32, 'CtxtHandle', CtxtHandle)
setattr(secur32, 'SecBuffer', SecBuffer)
setattr(secur32, 'SecBufferDesc', SecBufferDesc)
setattr(secur32, 'SecPkgContext_StreamSizes', SecPkgContext_StreamSizes)
setattr(secur32, 'SecPkgContext_ConnectionInfo', SecPkgContext_ConnectionInfo)
setattr(secur32, 'SCHANNEL_CRED', SCHANNEL_CRED)


def get_error():
    error = ctypes.GetLastError()
    return (error, ctypes.FormatError(error))



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

from ctypes import c_void_p, c_long, c_uint32, c_char_p, c_byte, c_ulong, c_bool
from ctypes import CDLL, string_at, cast, POINTER, byref
import ctypes

from .._ffi import FFIEngineError, buffer_from_bytes, byte_string_from_buffer


__all__ = [
    'CFHelpers',
    'CoreFoundation',
]


core_foundation_path = '/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation'

CoreFoundation = CDLL(core_foundation_path, use_errno=True)

CFIndex = c_long
CFStringEncoding = c_uint32
CFArray = c_void_p
CFData = c_void_p
CFString = c_void_p
CFNumber = c_void_p
CFDictionary = c_void_p
CFError = c_void_p
CFType = c_void_p
CFTypeID = c_ulong
CFBoolean = c_void_p
CFNumberType = c_uint32

CFTypeRef = POINTER(CFType)
CFArrayRef = POINTER(CFArray)
CFDataRef = POINTER(CFData)
CFStringRef = POINTER(CFString)
CFNumberRef = POINTER(CFNumber)
CFBooleanRef = POINTER(CFBoolean)
CFDictionaryRef = POINTER(CFDictionary)
CFErrorRef = POINTER(CFError)
CFAllocatorRef = c_void_p
CFDictionaryKeyCallBacks = c_void_p
CFDictionaryValueCallBacks = c_void_p
CFArrayCallBacks = c_void_p

pointer_p = POINTER(c_void_p)

try:
    CoreFoundation.CFDataGetLength.argtypes = [
        CFDataRef
    ]
    CoreFoundation.CFDataGetLength.restype = CFIndex

    CoreFoundation.CFDataGetBytePtr.argtypes = [
        CFDataRef
    ]
    CoreFoundation.CFDataGetBytePtr.restype = c_void_p

    CoreFoundation.CFDataCreate.argtypes = [
        CFAllocatorRef,
        c_char_p,
        CFIndex
    ]
    CoreFoundation.CFDataCreate.restype = CFDataRef

    CoreFoundation.CFDictionaryCreate.argtypes = [
        CFAllocatorRef,
        CFStringRef,
        CFTypeRef,
        CFIndex,
        CFDictionaryKeyCallBacks,
        CFDictionaryValueCallBacks
    ]
    CoreFoundation.CFDictionaryCreate.restype = CFDictionaryRef

    CoreFoundation.CFDictionaryGetCount.argtypes = [
        CFDictionaryRef
    ]
    CoreFoundation.CFDictionaryGetCount.restype = CFIndex

    CoreFoundation.CFStringGetCStringPtr.argtypes = [
        CFStringRef,
        CFStringEncoding
    ]
    CoreFoundation.CFStringGetCStringPtr.restype = c_char_p

    CoreFoundation.CFStringGetCString.argtypes = [
        CFStringRef,
        c_char_p,
        CFIndex,
        CFStringEncoding
    ]
    CoreFoundation.CFStringGetCString.restype = c_bool

    CoreFoundation.CFStringCreateWithCString.argtypes = [
        CFAllocatorRef,
        c_char_p,
        CFStringEncoding
    ]
    CoreFoundation.CFStringCreateWithCString.restype = CFStringRef

    CoreFoundation.CFNumberCreate.argtypes = [
        CFAllocatorRef,
        CFNumberType,
        c_void_p
    ]
    CoreFoundation.CFNumberCreate.restype = CFNumberRef

    CoreFoundation.CFCopyTypeIDDescription.argtypes = [
        CFTypeID
    ]
    CoreFoundation.CFCopyTypeIDDescription.restype = CFStringRef

    CoreFoundation.CFRelease.argtypes = [
        CFTypeRef
    ]
    CoreFoundation.CFRelease.restype = None

    CoreFoundation.CFRetain.argtypes = [
        CFTypeRef
    ]
    CoreFoundation.CFRetain.restype = None

    CoreFoundation.CFErrorCopyDescription.argtypes = [
        CFErrorRef
    ]
    CoreFoundation.CFErrorCopyDescription.restype = CFStringRef

    CoreFoundation.CFErrorGetDomain.argtypes = [
        CFErrorRef
    ]
    CoreFoundation.CFErrorGetDomain.restype = CFStringRef

    CoreFoundation.CFErrorGetCode.argtypes = [
        CFErrorRef
    ]
    CoreFoundation.CFErrorGetCode.restype = CFIndex

    CoreFoundation.CFBooleanGetValue.argtypes = [
        CFBooleanRef
    ]
    CoreFoundation.CFBooleanGetValue.restype = c_byte

    CoreFoundation.CFDictionaryGetTypeID.argtypes = []
    CoreFoundation.CFDictionaryGetTypeID.restype = CFTypeID

    CoreFoundation.CFNumberGetTypeID.argtypes = []
    CoreFoundation.CFNumberGetTypeID.restype = CFTypeID

    CoreFoundation.CFStringGetTypeID.argtypes = []
    CoreFoundation.CFStringGetTypeID.restype = CFTypeID

    CoreFoundation.CFDataGetTypeID.argtypes = []
    CoreFoundation.CFDataGetTypeID.restype = CFTypeID

    CoreFoundation.CFArrayCreate.argtypes = [
        CFAllocatorRef,
        POINTER(c_void_p),
        CFIndex,
        CFArrayCallBacks
    ]
    CoreFoundation.CFArrayCreate.restype = CFArrayRef

    CoreFoundation.CFArrayGetCount.argtypes = [
        CFArrayRef
    ]
    CoreFoundation.CFArrayGetCount.restype = CFIndex

    CoreFoundation.CFArrayGetValueAtIndex.argtypes = [
        CFArrayRef,
        CFIndex
    ]
    CoreFoundation.CFArrayGetValueAtIndex.restype = CFTypeRef

    CoreFoundation.CFNumberGetType.argtypes = [
        CFNumberRef
    ]
    CoreFoundation.CFNumberGetType.restype = CFNumberType

    CoreFoundation.CFNumberGetValue.argtypes = [
        CFNumberRef,
        CFNumberType,
        c_void_p
    ]
    CoreFoundation.CFNumberGetValue.restype = c_bool

    CoreFoundation.CFDictionaryGetKeysAndValues.argtypes = [
        CFDictionaryRef,
        pointer_p,
        pointer_p
    ]
    CoreFoundation.CFDictionaryGetKeysAndValues.restype = CFIndex

    CoreFoundation.CFGetTypeID.argtypes = [
        CFTypeRef
    ]
    CoreFoundation.CFGetTypeID.restype = CFTypeID

    setattr(CoreFoundation, 'kCFAllocatorDefault', CFAllocatorRef.in_dll(CoreFoundation, 'kCFAllocatorDefault'))
    setattr(CoreFoundation, 'kCFBooleanTrue', CFTypeRef.in_dll(CoreFoundation, 'kCFBooleanTrue'))

    kCFTypeDictionaryKeyCallBacks = c_void_p.in_dll(CoreFoundation, 'kCFTypeDictionaryKeyCallBacks')
    kCFTypeDictionaryValueCallBacks = c_void_p.in_dll(CoreFoundation, 'kCFTypeDictionaryValueCallBacks')
    kCFTypeArrayCallBacks = c_void_p.in_dll(CoreFoundation, 'kCFTypeArrayCallBacks')

except (AttributeError):
    raise FFIEngineError('Error initializing ctypes')

setattr(CoreFoundation, 'CFDataRef', CFDataRef)
setattr(CoreFoundation, 'CFErrorRef', CFErrorRef)
setattr(CoreFoundation, 'CFArrayRef', CFArrayRef)
kCFNumberCFIndexType = CFNumberType(14)
kCFStringEncodingUTF8 = CFStringEncoding(0x08000100)


def _cast_pointer_p(value):
    """
    Casts a value to a pointer of a pointer

    :param value:
        A ctypes object

    :return:
        A POINTER(c_void_p) object
    """

    return cast(value, pointer_p)


class CFHelpers():
    """
    Namespace for core foundation helpers
    """

    _native_map = {}

    @classmethod
    def register_native_mapping(cls, type_id, callback):
        """
        Register a function to convert a core foundation data type into its
        equivalent in python

        :param type_id:
            The CFTypeId for the type

        :param callback:
            A callback to pass the CFType object to
        """

        cls._native_map[int(type_id)] = callback

    @staticmethod
    def cf_number_to_number(value):
        """
        Converts a CFNumber object to a python float or integer

        :param value:
            The CFNumber object

        :return:
            A python number (float or integer)
        """

        type_ = CoreFoundation.CFNumberGetType(_cast_pointer_p(value))
        c_type = {
            1: c_byte,              # kCFNumberSInt8Type
            2: ctypes.c_short,      # kCFNumberSInt16Type
            3: ctypes.c_int32,      # kCFNumberSInt32Type
            4: ctypes.c_int64,      # kCFNumberSInt64Type
            5: ctypes.c_float,      # kCFNumberFloat32Type
            6: ctypes.c_double,     # kCFNumberFloat64Type
            7: c_byte,              # kCFNumberCharType
            8: ctypes.c_short,      # kCFNumberShortType
            9: ctypes.c_int,        # kCFNumberIntType
            10: c_long,             # kCFNumberLongType
            11: ctypes.c_longlong,  # kCFNumberLongLongType
            12: ctypes.c_float,     # kCFNumberFloatType
            13: ctypes.c_double,    # kCFNumberDoubleType
            14: c_long,             # kCFNumberCFIndexType
            15: ctypes.c_int,       # kCFNumberNSIntegerType
            16: ctypes.c_double,    # kCFNumberCGFloatType
        }[type_]
        output = c_type(0)
        CoreFoundation.CFNumberGetValue(_cast_pointer_p(value), type_, byref(output))
        return output.value

    @staticmethod
    def cf_dictionary_to_dict(dictionary):
        """
        Converts a CFDictionary object into a python dictionary

        :param dictionary:
            The CFDictionary to convert

        :return:
            A python dict
        """

        dict_length = CoreFoundation.CFDictionaryGetCount(dictionary)

        keys = (CFTypeRef * dict_length)()
        values = (CFTypeRef * dict_length)()
        CoreFoundation.CFDictionaryGetKeysAndValues(
            dictionary,
            _cast_pointer_p(keys),
            _cast_pointer_p(values)
        )

        output = {}
        for index in range(0, dict_length):
            output[CFHelpers.native(keys[index])] = CFHelpers.native(values[index])

        return output

    @classmethod
    def native(cls, value):
        """
        Converts a CF* object into its python equivalent

        :param value:
            The CF* object to convert

        :return:
            The native python object
        """

        type_id = CoreFoundation.CFGetTypeID(value)
        if type_id in cls._native_map:
            return cls._native_map[type_id](value)
        else:
            return value

    @staticmethod
    def cf_string_to_unicode(value):
        """
        Creates a python unicode string from a CFString object

        :param value:
            The CFString to convert

        :return:
            A python unicode string
        """

        string = CoreFoundation.CFStringGetCStringPtr(
            _cast_pointer_p(value),
            kCFStringEncodingUTF8
        )
        if string is None:
            buffer = buffer_from_bytes(1024)
            result = CoreFoundation.CFStringGetCString(
                _cast_pointer_p(value),
                buffer,
                1024,
                kCFStringEncodingUTF8
            )
            if not result:
                raise OSError('Error copying C string from CFStringRef')
            string = byte_string_from_buffer(buffer)
        if string is not None:
            string = string.decode('utf-8')
        return string

    @staticmethod
    def cf_string_from_unicode(string):
        """
        Creates a CFStringRef object from a unicode string

        :param string:
            The unicode string to create the CFString object from

        :return:
            A CFStringRef
        """

        return CoreFoundation.CFStringCreateWithCString(
            CoreFoundation.kCFAllocatorDefault,
            string.encode('utf-8'),
            kCFStringEncodingUTF8
        )

    @staticmethod
    def cf_data_to_bytes(value):
        """
        Extracts a bytestring from a CFData object

        :param value:
            A CFData object

        :return:
            A byte string
        """

        start = CoreFoundation.CFDataGetBytePtr(value)
        num_bytes = CoreFoundation.CFDataGetLength(value)
        return string_at(start, num_bytes)

    @staticmethod
    def cf_data_from_bytes(bytes_):
        """
        Creates a CFDataRef object from a byte string

        :param bytes_:
            The data to create the CFData object from

        :return:
            A CFDataRef
        """

        return CoreFoundation.CFDataCreate(
            CoreFoundation.kCFAllocatorDefault,
            bytes_,
            len(bytes_)
        )

    @staticmethod
    def cf_dictionary_from_pairs(pairs):
        """
        Creates a CFDictionaryRef object from a list of 2-element tuples
        representing the key and value. Each key should be a CFStringRef and each
        value some sort of CF* type.

        :param pairs:
            A list of 2-element tuples

        :return:
            A CFDictionaryRef
        """

        length = len(pairs)
        keys = []
        values = []
        for pair in pairs:
            key, value = pair
            keys.append(key)
            values.append(value)
        keys = (CFStringRef * length)(*keys)
        values = (CFTypeRef * length)(*values)
        return CoreFoundation.CFDictionaryCreate(
            CoreFoundation.kCFAllocatorDefault,
            _cast_pointer_p(byref(keys)),
            _cast_pointer_p(byref(values)),
            length,
            kCFTypeDictionaryKeyCallBacks,
            kCFTypeDictionaryValueCallBacks
        )

    @staticmethod
    def cf_array_from_list(values):
        """
        Creates a CFArrayRef object from a list of CF* type objects.

        :param values:
            A list of CF* type object

        :return:
            A CFArrayRef
        """

        length = len(values)
        values = (CFTypeRef * length)(*values)
        return CoreFoundation.CFArrayCreate(
            CoreFoundation.kCFAllocatorDefault,
            _cast_pointer_p(byref(values)),
            length,
            kCFTypeArrayCallBacks
        )

    @staticmethod
    def cf_number_from_integer(integer):
        """
        Creates a CFNumber object from an integer

        :param integer:
            The integer to create the CFNumber for

        :return:
            A CFNumber
        """

        integer_as_long = c_long(integer)
        return CoreFoundation.CFNumberCreate(
            CoreFoundation.kCFAllocatorDefault,
            kCFNumberCFIndexType,
            byref(integer_as_long)
        )



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import datetime
import hashlib
import struct

from .._asn1 import Certificate
from .._ffi import (
    array_from_pointer,
    buffer_from_bytes,
    bytes_from_buffer,
    cast,
    deref,
    is_null,
    new,
    null,
    struct_from_buffer,
    unwrap,
)
from ._crypt32 import crypt32, Crypt32Const, get_error, handle_error
from .._types import str_cls


__all__ = [
    'extract_from_system',
    'system_path',
]


def system_path():
    return None


def extract_from_system(cert_callback=None, callback_only_on_failure=False):
    """
    Extracts trusted CA certificates from the Windows certificate store

    :param cert_callback:
        A callback that is called once for each certificate in the trust store.
        It should accept two parameters: an asn1crypto.x509.Certificate object,
        and a reason. The reason will be None if the certificate is being
        exported, otherwise it will be a unicode string of the reason it won't.

    :param callback_only_on_failure:
        A boolean - if the callback should only be called when a certificate is
        not exported.

    :raises:
        OSError - when an error is returned by the OS crypto library

    :return:
        A list of 3-element tuples:
         - 0: a byte string of a DER-encoded certificate
         - 1: a set of unicode strings that are OIDs of purposes to trust the
              certificate for
         - 2: a set of unicode strings that are OIDs of purposes to reject the
              certificate for
    """

    certificates = {}
    processed = {}

    now = datetime.datetime.utcnow()

    for store in ["ROOT", "CA"]:
        store_handle = crypt32.CertOpenSystemStoreW(null(), store)
        handle_error(store_handle)

        context_pointer = null()
        while True:
            context_pointer = crypt32.CertEnumCertificatesInStore(store_handle, context_pointer)
            if is_null(context_pointer):
                break
            context = unwrap(context_pointer)

            trust_all = False
            data = None
            digest = None

            if context.dwCertEncodingType != Crypt32Const.X509_ASN_ENCODING:
                continue

            data = bytes_from_buffer(context.pbCertEncoded, int(context.cbCertEncoded))
            digest = hashlib.sha1(data).digest()
            if digest in processed:
                continue

            processed[digest] = True
            cert_info = unwrap(context.pCertInfo)

            not_before_seconds = _convert_filetime_to_timestamp(cert_info.NotBefore)
            try:
                not_before = datetime.datetime.fromtimestamp(not_before_seconds)
                if not_before > now:
                    if cert_callback:
                        cert_callback(Certificate.load(data), 'not yet valid')
                    continue
            except (ValueError, OSError):
                # If there is an error converting the not before timestamp,
                # it is almost certainly because it is from too long ago,
                # which means the cert is definitely valid by now.
                pass

            not_after_seconds = _convert_filetime_to_timestamp(cert_info.NotAfter)
            try:
                not_after = datetime.datetime.fromtimestamp(not_after_seconds)
                if not_after < now:
                    if cert_callback:
                        cert_callback(Certificate.load(data), 'no longer valid')
                    continue
            except (ValueError, OSError) as e:
                # The only reason we would get an exception here is if the
                # expiration time is so far in the future that it can't be
                # used as a timestamp, or it is before 0. If it is very far
                # in the future, the cert is still valid, so we only raise
                # an exception if the timestamp is less than zero.
                if not_after_seconds < 0:
                    message = e.args[0] + ' - ' + str_cls(not_after_seconds)
                    e.args = (message,) + e.args[1:]
                    raise e

            trust_oids = set()
            reject_oids = set()

            # Here we grab the extended key usage properties that Windows
            # layers on top of the extended key usage extension that is
            # part of the certificate itself. For highest security, users
            # should only use certificates for the intersection of the two
            # lists of purposes. However, many seen to treat the OS trust
            # list as an override.
            to_read = new(crypt32, 'DWORD *', 0)
            res = crypt32.CertGetEnhancedKeyUsage(
                context_pointer,
                Crypt32Const.CERT_FIND_PROP_ONLY_ENHKEY_USAGE_FLAG,
                null(),
                to_read
            )

            # Per the Microsoft documentation, if CRYPT_E_NOT_FOUND is returned
            # from get_error(), it means the certificate is valid for all purposes
            error_code, _ = get_error()
            if not res and error_code != Crypt32Const.CRYPT_E_NOT_FOUND:
                handle_error(res)

            if error_code == Crypt32Const.CRYPT_E_NOT_FOUND:
                trust_all = True
            else:
                usage_buffer = buffer_from_bytes(deref(to_read))
                res = crypt32.CertGetEnhancedKeyUsage(
                    context_pointer,
                    Crypt32Const.CERT_FIND_PROP_ONLY_ENHKEY_USAGE_FLAG,
                    cast(crypt32, 'CERT_ENHKEY_USAGE *', usage_buffer),
                    to_read
                )
                handle_error(res)

                key_usage_pointer = struct_from_buffer(crypt32, 'CERT_ENHKEY_USAGE', usage_buffer)
                key_usage = unwrap(key_usage_pointer)

                # Having no enhanced usage properties means a cert is distrusted
                if key_usage.cUsageIdentifier == 0:
                    if cert_callback:
                        cert_callback(Certificate.load(data), 'explicitly distrusted')
                    continue

                oids = array_from_pointer(
                    crypt32,
                    'LPCSTR',
                    key_usage.rgpszUsageIdentifier,
                    key_usage.cUsageIdentifier
                )
                for oid in oids:
                    trust_oids.add(oid.decode('ascii'))

            cert = None

            # If the certificate is not under blanket trust, we have to
            # determine what purposes it is rejected for by diffing the
            # set of OIDs from the certificate with the OIDs that are
            # trusted.
            if not trust_all:
                cert = Certificate.load(data)
                if cert.extended_key_usage_value:
                    for cert_oid in cert.extended_key_usage_value:
                        oid = cert_oid.dotted
                        if oid not in trust_oids:
                            reject_oids.add(oid)

            if cert_callback and not callback_only_on_failure:
                if cert is None:
                    cert = Certificate.load(data)
                cert_callback(cert, None)

            certificates[digest] = (data, trust_oids, reject_oids)

        result = crypt32.CertCloseStore(store_handle, 0)
        handle_error(result)
        store_handle = None

    return certificates.values()


def _convert_filetime_to_timestamp(filetime):
    """
    Windows returns times as 64-bit unsigned longs that are the number
    of hundreds of nanoseconds since Jan 1 1601. This converts it to
    a datetime object.

    :param filetime:
        A FILETIME struct object

    :return:
        An integer unix timestamp
    """

    hundreds_nano_seconds = struct.unpack(
        b'>Q',
        struct.pack(
            b'>LL',
            filetime.dwHighDateTime,
            filetime.dwLowDateTime
        )
    )[0]
    seconds_since_1601 = hundreds_nano_seconds / 10000000
    return seconds_since_1601 - 11644473600  # Seconds from Jan 1 1601 to Jan 1 1970



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

from .. import ffi
from .._ffi import is_null, unwrap

if ffi() == 'cffi':
    from ._core_foundation_cffi import CoreFoundation, CFHelpers
else:
    from ._core_foundation_ctypes import CoreFoundation, CFHelpers


_all__ = [
    'CFHelpers',
    'CoreFoundation',
    'handle_cf_error',
]


def handle_cf_error(error_pointer):
    """
    Checks a CFErrorRef and throws an exception if there is an error to report

    :param error_pointer:
        A CFErrorRef

    :raises:
        OSError - when the CFErrorRef contains an error
    """

    if is_null(error_pointer):
        return

    error = unwrap(error_pointer)
    if is_null(error):
        return

    cf_string_domain = CoreFoundation.CFErrorGetDomain(error)
    domain = CFHelpers.cf_string_to_unicode(cf_string_domain)
    CoreFoundation.CFRelease(cf_string_domain)
    num = CoreFoundation.CFErrorGetCode(error)

    cf_string_ref = CoreFoundation.CFErrorCopyDescription(error)
    output = CFHelpers.cf_string_to_unicode(cf_string_ref)
    CoreFoundation.CFRelease(cf_string_ref)

    if output is None:
        if domain == 'NSOSStatusErrorDomain':
            code_map = {
                -2147416010: 'ACL add failed',
                -2147416025: 'ACL base certs not supported',
                -2147416019: 'ACL challenge callback failed',
                -2147416015: 'ACL change failed',
                -2147416012: 'ACL delete failed',
                -2147416017: 'ACL entry tag not found',
                -2147416011: 'ACL replace failed',
                -2147416021: 'ACL subject type not supported',
                -2147415789: 'Algid mismatch',
                -2147415726: 'Already logged in',
                -2147415040: 'Apple add application ACL subject',
                -2147415036: 'Apple invalid key end date',
                -2147415037: 'Apple invalid key start date',
                -2147415039: 'Apple public key incomplete',
                -2147415038: 'Apple signature mismatch',
                -2147415034: 'Apple SSLv2 rollback',
                -2147415802: 'Attach handle busy',
                -2147415731: 'Block size mismatch',
                -2147415722: 'Crypto data callback failed',
                -2147415804: 'Device error',
                -2147415835: 'Device failed',
                -2147415803: 'Device memory error',
                -2147415836: 'Device reset',
                -2147415728: 'Device verify failed',
                -2147416054: 'Function failed',
                -2147416057: 'Function not implemented',
                -2147415807: 'Input length error',
                -2147415837: 'Insufficient client identification',
                -2147416063: 'Internal error',
                -2147416027: 'Invalid access credentials',
                -2147416026: 'Invalid ACL base certs',
                -2147416020: 'Invalid ACL challenge callback',
                -2147416016: 'Invalid ACL edit mode',
                -2147416018: 'Invalid ACL entry tag',
                -2147416022: 'Invalid ACL subject value',
                -2147415759: 'Invalid algorithm',
                -2147415678: 'Invalid attr access credentials',
                -2147415704: 'Invalid attr alg params',
                -2147415686: 'Invalid attr base',
                -2147415738: 'Invalid attr block size',
                -2147415680: 'Invalid attr dl db handle',
                -2147415696: 'Invalid attr effective bits',
                -2147415692: 'Invalid attr end date',
                -2147415752: 'Invalid attr init vector',
                -2147415682: 'Invalid attr iteration count',
                -2147415754: 'Invalid attr key',
                -2147415740: 'Invalid attr key length',
                -2147415700: 'Invalid attr key type',
                -2147415702: 'Invalid attr label',
                -2147415698: 'Invalid attr mode',
                -2147415708: 'Invalid attr output size',
                -2147415748: 'Invalid attr padding',
                -2147415742: 'Invalid attr passphrase',
                -2147415688: 'Invalid attr prime',
                -2147415674: 'Invalid attr private key format',
                -2147415676: 'Invalid attr public key format',
                -2147415746: 'Invalid attr random',
                -2147415706: 'Invalid attr rounds',
                -2147415750: 'Invalid attr salt',
                -2147415744: 'Invalid attr seed',
                -2147415694: 'Invalid attr start date',
                -2147415684: 'Invalid attr subprime',
                -2147415672: 'Invalid attr symmetric key format',
                -2147415690: 'Invalid attr version',
                -2147415670: 'Invalid attr wrapped key format',
                -2147415760: 'Invalid context',
                -2147416000: 'Invalid context handle',
                -2147415976: 'Invalid crypto data',
                -2147415994: 'Invalid data',
                -2147415768: 'Invalid data count',
                -2147415723: 'Invalid digest algorithm',
                -2147416059: 'Invalid input pointer',
                -2147415766: 'Invalid input vector',
                -2147415792: 'Invalid key',
                -2147415780: 'Invalid keyattr mask',
                -2147415782: 'Invalid keyusage mask',
                -2147415790: 'Invalid key class',
                -2147415776: 'Invalid key format',
                -2147415778: 'Invalid key label',
                -2147415783: 'Invalid key pointer',
                -2147415791: 'Invalid key reference',
                -2147415727: 'Invalid login name',
                -2147416014: 'Invalid new ACL entry',
                -2147416013: 'Invalid new ACL owner',
                -2147416058: 'Invalid output pointer',
                -2147415765: 'Invalid output vector',
                -2147415978: 'Invalid passthrough id',
                -2147416060: 'Invalid pointer',
                -2147416024: 'Invalid sample value',
                -2147415733: 'Invalid signature',
                -2147415787: 'Key blob type incorrect',
                -2147415786: 'Key header inconsistent',
                -2147415724: 'Key label already exists',
                -2147415788: 'Key usage incorrect',
                -2147416061: 'Mds error',
                -2147416062: 'Memory error',
                -2147415677: 'Missing attr access credentials',
                -2147415703: 'Missing attr alg params',
                -2147415685: 'Missing attr base',
                -2147415737: 'Missing attr block size',
                -2147415679: 'Missing attr dl db handle',
                -2147415695: 'Missing attr effective bits',
                -2147415691: 'Missing attr end date',
                -2147415751: 'Missing attr init vector',
                -2147415681: 'Missing attr iteration count',
                -2147415753: 'Missing attr key',
                -2147415739: 'Missing attr key length',
                -2147415699: 'Missing attr key type',
                -2147415701: 'Missing attr label',
                -2147415697: 'Missing attr mode',
                -2147415707: 'Missing attr output size',
                -2147415747: 'Missing attr padding',
                -2147415741: 'Missing attr passphrase',
                -2147415687: 'Missing attr prime',
                -2147415673: 'Missing attr private key format',
                -2147415675: 'Missing attr public key format',
                -2147415745: 'Missing attr random',
                -2147415705: 'Missing attr rounds',
                -2147415749: 'Missing attr salt',
                -2147415743: 'Missing attr seed',
                -2147415693: 'Missing attr start date',
                -2147415683: 'Missing attr subprime',
                -2147415671: 'Missing attr symmetric key format',
                -2147415689: 'Missing attr version',
                -2147415669: 'Missing attr wrapped key format',
                -2147415801: 'Not logged in',
                -2147415840: 'No user interaction',
                -2147416029: 'Object ACL not supported',
                -2147416028: 'Object ACL required',
                -2147416030: 'Object manip auth denied',
                -2147416031: 'Object use auth denied',
                -2147416032: 'Operation auth denied',
                -2147416055: 'OS access denied',
                -2147415806: 'Output length error',
                -2147415725: 'Private key already exists',
                -2147415730: 'Private key not found',
                -2147415989: 'Privilege not granted',
                -2147415805: 'Privilege not supported',
                -2147415729: 'Public key inconsistent',
                -2147415732: 'Query size unknown',
                -2147416023: 'Sample value not supported',
                -2147416056: 'Self check failed',
                -2147415838: 'Service not available',
                -2147415736: 'Staged operation in progress',
                -2147415735: 'Staged operation not started',
                -2147415779: 'Unsupported keyattr mask',
                -2147415781: 'Unsupported keyusage mask',
                -2147415785: 'Unsupported key format',
                -2147415777: 'Unsupported key label',
                -2147415784: 'Unsupported key size',
                -2147415839: 'User canceled',
                -2147415767: 'Vector of bufs unsupported',
                -2147415734: 'Verify failed',
            }
            if num in code_map:
                output = code_map[num]

        if not output:
            output = '%s %s' % (domain, num)

    raise OSError(output)


CFHelpers.register_native_mapping(
    CoreFoundation.CFStringGetTypeID(),
    CFHelpers.cf_string_to_unicode
)
CFHelpers.register_native_mapping(
    CoreFoundation.CFNumberGetTypeID(),
    CFHelpers.cf_number_to_number
)
CFHelpers.register_native_mapping(
    CoreFoundation.CFDataGetTypeID(),
    CFHelpers.cf_data_to_bytes
)
CFHelpers.register_native_mapping(
    CoreFoundation.CFDictionaryGetTypeID(),
    CFHelpers.cf_dictionary_to_dict
)



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import os

from ._errors import pretty_message
from ._types import type_name, int_types


__all__ = [
    'rand_bytes',
]


def rand_bytes(length):
    """
    Returns a number of random bytes suitable for cryptographic purposes

    :param length:
        The desired number of bytes

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by OpenSSL

    :return:
        A byte string
    """

    if not isinstance(length, int_types):
        raise TypeError(pretty_message(
            '''
            length must be an integer, not %s
            ''',
            type_name(length)
        ))

    if length < 1:
        raise ValueError('length must be greater than 0')

    if length > 1024:
        raise ValueError('length must not be greater than 1024')

    return os.urandom(length)



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import sys

import ctypes
from ctypes import windll, wintypes, POINTER, Structure, c_void_p, c_char_p
from ctypes.wintypes import DWORD

from .._ffi import FFIEngineError
from .._types import str_cls
from ..errors import LibraryNotFoundError
from ._kernel32 import kernel32


__all__ = [
    'crypt32',
    'get_error',
]


try:
    crypt32 = windll.crypt32
except (OSError) as e:
    if str_cls(e).find('The specified module could not be found') != -1:
        raise LibraryNotFoundError('crypt32.dll could not be found')
    raise

HCERTSTORE = wintypes.HANDLE
HCERTCHAINENGINE = wintypes.HANDLE
HCRYPTPROV = wintypes.HANDLE
HCRYPTKEY = wintypes.HANDLE
PBYTE = c_char_p
if sys.maxsize > 2 ** 32:
    ULONG_PTR = ctypes.c_uint64
else:
    ULONG_PTR = ctypes.c_ulong

try:
    class CRYPTOAPI_BLOB(Structure):  # noqa
        _fields_ = [
            ("cbData", DWORD),
            ("pbData", c_void_p),
        ]
    CRYPT_INTEGER_BLOB = CRYPTOAPI_BLOB
    CERT_NAME_BLOB = CRYPTOAPI_BLOB
    CRYPT_BIT_BLOB = CRYPTOAPI_BLOB
    CRYPT_OBJID_BLOB = CRYPTOAPI_BLOB

    class CRYPT_ALGORITHM_IDENTIFIER(Structure):  # noqa
        _fields_ = [
            ("pszObjId", wintypes.LPSTR),
            ("Parameters", CRYPT_OBJID_BLOB),
        ]

    class CERT_PUBLIC_KEY_INFO(Structure):  # noqa
        _fields_ = [
            ("Algorithm", CRYPT_ALGORITHM_IDENTIFIER),
            ("PublicKey", CRYPT_BIT_BLOB),
        ]

    class CERT_EXTENSION(Structure):  # noqa
        _fields_ = [
            ("pszObjId", wintypes.LPSTR),
            ("fCritical", wintypes.BOOL),
            ("Value", CRYPT_OBJID_BLOB),
        ]
    PCERT_EXTENSION = POINTER(CERT_EXTENSION)

    class CERT_INFO(Structure):  # noqa
        _fields_ = [
            ("dwVersion", DWORD),
            ("SerialNumber", CRYPT_INTEGER_BLOB),
            ("SignatureAlgorithm", CRYPT_ALGORITHM_IDENTIFIER),
            ("Issuer", CERT_NAME_BLOB),
            ("NotBefore", kernel32.FILETIME),
            ("NotAfter", kernel32.FILETIME),
            ("Subject", CERT_NAME_BLOB),
            ("SubjectPublicKeyInfo", CERT_PUBLIC_KEY_INFO),
            ("IssuerUniqueId", CRYPT_BIT_BLOB),
            ("SubjectUniqueId", CRYPT_BIT_BLOB),
            ("cExtension", DWORD),
            ("rgExtension", POINTER(PCERT_EXTENSION)),
        ]
    PCERT_INFO = POINTER(CERT_INFO)

    class CERT_CONTEXT(Structure):  # noqa
        _fields_ = [
            ("dwCertEncodingType", DWORD),
            ("pbCertEncoded", c_void_p),
            ("cbCertEncoded", DWORD),
            ("pCertInfo", PCERT_INFO),
            ("hCertStore", HCERTSTORE)
        ]

    PCERT_CONTEXT = POINTER(CERT_CONTEXT)

    class CERT_ENHKEY_USAGE(Structure):  # noqa
        _fields_ = [
            ('cUsageIdentifier', DWORD),
            ('rgpszUsageIdentifier', POINTER(POINTER(wintypes.BYTE))),
        ]

    PCERT_ENHKEY_USAGE = POINTER(CERT_ENHKEY_USAGE)

    class CERT_TRUST_STATUS(Structure):  # noqa
        _fields_ = [
            ('dwErrorStatus', DWORD),
            ('dwInfoStatus', DWORD),
        ]

    class CERT_CHAIN_ELEMENT(Structure):  # noqa
        _fields_ = [
            ('cbSize', DWORD),
            ('pCertContext', PCERT_CONTEXT),
            ('TrustStatus', CERT_TRUST_STATUS),
            ('pRevocationInfo', c_void_p),
            ('pIssuanceUsage', PCERT_ENHKEY_USAGE),
            ('pApplicationUsage', PCERT_ENHKEY_USAGE),
            ('pwszExtendedErrorInfo', wintypes.LPCWSTR),
        ]

    PCERT_CHAIN_ELEMENT = POINTER(CERT_CHAIN_ELEMENT)

    class CERT_SIMPLE_CHAIN(Structure):  # noqa
        _fields_ = [
            ('cbSize', DWORD),
            ('TrustStatus', CERT_TRUST_STATUS),
            ('cElement', DWORD),
            ('rgpElement', POINTER(PCERT_CHAIN_ELEMENT)),
            ('pTrustListInfo', c_void_p),
            ('fHasRevocationFreshnessTime', wintypes.BOOL),
            ('dwRevocationFreshnessTime', DWORD),
        ]

    PCERT_SIMPLE_CHAIN = POINTER(CERT_SIMPLE_CHAIN)

    class CERT_CHAIN_CONTEXT(Structure):  # noqa
        _fields_ = [
            ('cbSize', DWORD),
            ('TrustStatus', CERT_TRUST_STATUS),
            ('cChain', DWORD),
            ('rgpChain', POINTER(PCERT_SIMPLE_CHAIN)),
            ('cLowerQualityChainContext', DWORD),
            ('rgpLowerQualityChainContext', c_void_p),
            ('fHasRevocationFreshnessTime', wintypes.BOOL),
            ('dwRevocationFreshnessTime', DWORD),
        ]

    PCERT_CHAIN_CONTEXT = POINTER(CERT_CHAIN_CONTEXT)

    class CERT_USAGE_MATCH(Structure):  # noqa
        _fields_ = [
            ('dwType', DWORD),
            ('Usage', CERT_ENHKEY_USAGE),
        ]

    class CERT_CHAIN_PARA(Structure):  # noqa
        _fields_ = [
            ('cbSize', DWORD),
            ('RequestedUsage', CERT_USAGE_MATCH),
        ]

    class CERT_CHAIN_POLICY_PARA(Structure):  # noqa
        _fields_ = [
            ('cbSize', DWORD),
            ('dwFlags', DWORD),
            ('pvExtraPolicyPara', c_void_p),
        ]

    class SSL_EXTRA_CERT_CHAIN_POLICY_PARA(Structure):  # noqa
        _fields_ = [
            ('cbSize', DWORD),
            ('dwAuthType', DWORD),
            ('fdwChecks', DWORD),
            ('pwszServerName', wintypes.LPCWSTR),
        ]

    class CERT_CHAIN_POLICY_STATUS(Structure):  # noqa
        _fields_ = [
            ('cbSize', DWORD),
            ('dwError', DWORD),
            ('lChainIndex', wintypes.LONG),
            ('lElementIndex', wintypes.LONG),
            ('pvExtraPolicyStatus', c_void_p),
        ]

    crypt32.CertOpenStore.argtypes = [
        wintypes.LPCSTR,
        DWORD,
        HCRYPTPROV,
        DWORD,
        c_void_p
    ]
    crypt32.CertOpenStore.restype = HCERTSTORE

    crypt32.CertAddEncodedCertificateToStore.argtypes = [
        HCERTSTORE,
        DWORD,
        PBYTE,
        DWORD,
        DWORD,
        POINTER(PCERT_CONTEXT)
    ]
    crypt32.CertAddEncodedCertificateToStore.restype = wintypes.BOOL

    crypt32.CertGetCertificateChain.argtypes = [
        HCERTCHAINENGINE,
        PCERT_CONTEXT,
        POINTER(kernel32.FILETIME),
        HCERTSTORE,
        POINTER(CERT_CHAIN_PARA),
        DWORD,
        c_void_p,
        POINTER(PCERT_CHAIN_CONTEXT)
    ]
    crypt32.CertGetCertificateChain.restype = wintypes.BOOL

    crypt32.CertVerifyCertificateChainPolicy.argtypes = [
        ULONG_PTR,
        PCERT_CHAIN_CONTEXT,
        POINTER(CERT_CHAIN_POLICY_PARA),
        POINTER(CERT_CHAIN_POLICY_STATUS)
    ]
    crypt32.CertVerifyCertificateChainPolicy.restype = wintypes.BOOL

    crypt32.CertFreeCertificateChain.argtypes = [
        PCERT_CHAIN_CONTEXT
    ]
    crypt32.CertFreeCertificateChain.restype = None

    crypt32.CertOpenSystemStoreW.argtypes = [
        wintypes.HANDLE,
        wintypes.LPCWSTR
    ]
    crypt32.CertOpenSystemStoreW.restype = HCERTSTORE

    crypt32.CertEnumCertificatesInStore.argtypes = [
        HCERTSTORE,
        PCERT_CONTEXT
    ]
    crypt32.CertEnumCertificatesInStore.restype = PCERT_CONTEXT

    crypt32.CertCloseStore.argtypes = [
        HCERTSTORE,
        DWORD
    ]
    crypt32.CertCloseStore.restype = wintypes.BOOL

    crypt32.CertGetEnhancedKeyUsage.argtypes = [
        PCERT_CONTEXT,
        DWORD,
        c_void_p,
        POINTER(DWORD)
    ]
    crypt32.CertGetEnhancedKeyUsage.restype = wintypes.BOOL

except (AttributeError):
    raise FFIEngineError('Error initializing ctypes')


setattr(crypt32, 'FILETIME', kernel32.FILETIME)
setattr(crypt32, 'CERT_ENHKEY_USAGE', CERT_ENHKEY_USAGE)
setattr(crypt32, 'CERT_CONTEXT', CERT_CONTEXT)
setattr(crypt32, 'PCERT_CONTEXT', PCERT_CONTEXT)
setattr(crypt32, 'CERT_USAGE_MATCH', CERT_USAGE_MATCH)
setattr(crypt32, 'CERT_CHAIN_PARA', CERT_CHAIN_PARA)
setattr(crypt32, 'CERT_CHAIN_POLICY_PARA', CERT_CHAIN_POLICY_PARA)
setattr(crypt32, 'SSL_EXTRA_CERT_CHAIN_POLICY_PARA', SSL_EXTRA_CERT_CHAIN_POLICY_PARA)
setattr(crypt32, 'CERT_CHAIN_POLICY_STATUS', CERT_CHAIN_POLICY_STATUS)
setattr(crypt32, 'PCERT_CHAIN_CONTEXT', PCERT_CHAIN_CONTEXT)


def get_error():
    error = ctypes.GetLastError()
    return (error, ctypes.FormatError(error))



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import sys
import re
import socket as socket_
import select
import numbers

from ._libssl import libssl, LibsslConst
from ._libcrypto import libcrypto, libcrypto_version_info, handle_openssl_error, peek_openssl_error
from .. import _backend_config
from .._asn1 import Certificate as Asn1Certificate
from .._errors import pretty_message
from .._ffi import null, bytes_from_buffer, buffer_from_bytes, is_null, buffer_pointer
from .._types import type_name, str_cls, byte_cls, int_types
from ..errors import TLSError, TLSDisconnectError, TLSGracefulDisconnectError
from .._tls import (
    detect_client_auth_request,
    extract_chain,
    get_dh_params_length,
    parse_session_info,
    raise_client_auth,
    raise_dh_params,
    raise_disconnection,
    raise_expired_not_yet_valid,
    raise_handshake,
    raise_hostname,
    raise_no_issuer,
    raise_protocol_error,
    raise_protocol_version,
    raise_self_signed,
    raise_verification,
    raise_weak_signature,
    parse_tls_records,
    parse_handshake_messages,
)
from .asymmetric import load_certificate, Certificate
from ..keys import parse_certificate
from ..trust_list import get_path

if sys.version_info < (3,):
    range = xrange  # noqa

if sys.version_info < (3, 7):
    Pattern = re._pattern_type
else:
    Pattern = re.Pattern


__all__ = [
    'TLSSession',
    'TLSSocket',
]


_trust_list_path = _backend_config().get('trust_list_path')
_line_regex = re.compile(b'(\r\n|\r|\n)')
_PROTOCOL_MAP = {
    'SSLv2': LibsslConst.SSL_OP_NO_SSLv2,
    'SSLv3': LibsslConst.SSL_OP_NO_SSLv3,
    'TLSv1': LibsslConst.SSL_OP_NO_TLSv1,
    'TLSv1.1': LibsslConst.SSL_OP_NO_TLSv1_1,
    'TLSv1.2': LibsslConst.SSL_OP_NO_TLSv1_2,
}


def _homogenize_openssl3_error(error_tuple):
    """
    Takes a 3-element tuple from peek_openssl_error() and modifies it
    to handle the changes in OpenSSL 3.0. That release removed the
    concept of an error function, meaning the second item in the tuple
    will always be 0.

    :param error_tuple:
        A 3-element tuple of integers

    :return:
        A 3-element tuple of integers
    """

    if libcrypto_version_info < (3,):
        return error_tuple
    return (error_tuple[0], 0, error_tuple[2])


class TLSSession(object):
    """
    A TLS session object that multiple TLSSocket objects can share for the
    sake of session reuse
    """

    _protocols = None
    _ciphers = None
    _manual_validation = None
    _extra_trust_roots = None
    _ssl_ctx = None
    _ssl_session = None

    def __init__(self, protocol=None, manual_validation=False, extra_trust_roots=None):
        """
        :param protocol:
            A unicode string or set of unicode strings representing allowable
            protocols to negotiate with the server:

             - "TLSv1.2"
             - "TLSv1.1"
             - "TLSv1"
             - "SSLv3"

            Default is: {"TLSv1", "TLSv1.1", "TLSv1.2"}

        :param manual_validation:
            If certificate and certificate path validation should be skipped
            and left to the developer to implement

        :param extra_trust_roots:
            A list containing one or more certificates to be treated as trust
            roots, in one of the following formats:
             - A byte string of the DER encoded certificate
             - A unicode string of the certificate filename
             - An asn1crypto.x509.Certificate object
             - An oscrypto.asymmetric.Certificate object

        :raises:
            ValueError - when any of the parameters contain an invalid value
            TypeError - when any of the parameters are of the wrong type
            OSError - when an error is returned by the OS crypto library
        """

        if not isinstance(manual_validation, bool):
            raise TypeError(pretty_message(
                '''
                manual_validation must be a boolean, not %s
                ''',
                type_name(manual_validation)
            ))

        self._manual_validation = manual_validation

        if protocol is None:
            protocol = set(['TLSv1', 'TLSv1.1', 'TLSv1.2'])

        if isinstance(protocol, str_cls):
            protocol = set([protocol])
        elif not isinstance(protocol, set):
            raise TypeError(pretty_message(
                '''
                protocol must be a unicode string or set of unicode strings,
                not %s
                ''',
                type_name(protocol)
            ))

        valid_protocols = set(['SSLv3', 'TLSv1', 'TLSv1.1', 'TLSv1.2'])
        unsupported_protocols = protocol - valid_protocols
        if unsupported_protocols:
            raise ValueError(pretty_message(
                '''
                protocol must contain only the unicode strings "SSLv3", "TLSv1",
                "TLSv1.1", "TLSv1.2", not %s
                ''',
                repr(unsupported_protocols)
            ))

        self._protocols = protocol

        self._extra_trust_roots = []
        if extra_trust_roots:
            for extra_trust_root in extra_trust_roots:
                if isinstance(extra_trust_root, Certificate):
                    extra_trust_root = extra_trust_root.asn1
                elif isinstance(extra_trust_root, byte_cls):
                    extra_trust_root = parse_certificate(extra_trust_root)
                elif isinstance(extra_trust_root, str_cls):
                    with open(extra_trust_root, 'rb') as f:
                        extra_trust_root = parse_certificate(f.read())
                elif not isinstance(extra_trust_root, Asn1Certificate):
                    raise TypeError(pretty_message(
                        '''
                        extra_trust_roots must be a list of byte strings, unicode
                        strings, asn1crypto.x509.Certificate objects or
                        oscrypto.asymmetric.Certificate objects, not %s
                        ''',
                        type_name(extra_trust_root)
                    ))
                self._extra_trust_roots.append(extra_trust_root)

        ssl_ctx = None
        try:
            if libcrypto_version_info < (1, 1):
                method = libssl.SSLv23_method()
            else:
                method = libssl.TLS_method()
            ssl_ctx = libssl.SSL_CTX_new(method)
            if is_null(ssl_ctx):
                handle_openssl_error(0)
            self._ssl_ctx = ssl_ctx

            libssl.SSL_CTX_set_timeout(ssl_ctx, 600)

            # Allow caching SSL sessions
            libssl.SSL_CTX_ctrl(
                ssl_ctx,
                LibsslConst.SSL_CTRL_SET_SESS_CACHE_MODE,
                LibsslConst.SSL_SESS_CACHE_CLIENT,
                null()
            )

            if sys.platform in set(['win32', 'darwin']):
                trust_list_path = _trust_list_path
                if trust_list_path is None:
                    trust_list_path = get_path()

                if sys.platform == 'win32':
                    path_encoding = 'mbcs'
                else:
                    path_encoding = 'utf-8'
                result = libssl.SSL_CTX_load_verify_locations(
                    ssl_ctx,
                    trust_list_path.encode(path_encoding),
                    null()
                )

            else:
                result = libssl.SSL_CTX_set_default_verify_paths(ssl_ctx)
            handle_openssl_error(result)

            verify_mode = LibsslConst.SSL_VERIFY_NONE if manual_validation else LibsslConst.SSL_VERIFY_PEER
            libssl.SSL_CTX_set_verify(ssl_ctx, verify_mode, null())

            # Modern cipher suite list from https://wiki.mozilla.org/Security/Server_Side_TLS late August 2015
            result = libssl.SSL_CTX_set_cipher_list(
                ssl_ctx,
                (
                    b'ECDHE-RSA-AES128-GCM-SHA256:ECDHE-ECDSA-AES128-GCM-SHA256:'
                    b'ECDHE-RSA-AES256-GCM-SHA384:ECDHE-ECDSA-AES256-GCM-SHA384:'
                    b'DHE-RSA-AES128-GCM-SHA256:DHE-DSS-AES128-GCM-SHA256:'
                    b'kEDH+AESGCM:ECDHE-RSA-AES128-SHA256:ECDHE-ECDSA-AES128-SHA256:'
                    b'ECDHE-RSA-AES128-SHA:ECDHE-ECDSA-AES128-SHA:ECDHE-RSA-AES256-SHA384:'
                    b'ECDHE-ECDSA-AES256-SHA384:ECDHE-RSA-AES256-SHA:ECDHE-ECDSA-AES256-SHA:'
                    b'DHE-RSA-AES128-SHA256:DHE-RSA-AES128-SHA:DHE-DSS-AES128-SHA256:'
                    b'DHE-RSA-AES256-SHA256:DHE-DSS-AES256-SHA:DHE-RSA-AES256-SHA:'
                    b'AES128-GCM-SHA256:AES256-GCM-SHA384:AES128-SHA256:AES256-SHA256:'
                    b'AES128-SHA:AES256-SHA:AES:CAMELLIA:DES-CBC3-SHA:!aNULL:!eNULL:'
                    b'!EXPORT:!DES:!RC4:!MD5:!PSK:!aECDH:!EDH-DSS-DES-CBC3-SHA:'
                    b'!EDH-RSA-DES-CBC3-SHA:!KRB5-DES-CBC3-SHA'
                )
            )
            handle_openssl_error(result)

            disabled_protocols = set(['SSLv2'])
            disabled_protocols |= (valid_protocols - self._protocols)
            for disabled_protocol in disabled_protocols:
                libssl.SSL_CTX_ctrl(
                    ssl_ctx,
                    LibsslConst.SSL_CTRL_OPTIONS,
                    _PROTOCOL_MAP[disabled_protocol],
                    null()
                )

            if self._extra_trust_roots:
                x509_store = libssl.SSL_CTX_get_cert_store(ssl_ctx)
                for cert in self._extra_trust_roots:
                    oscrypto_cert = load_certificate(cert)
                    result = libssl.X509_STORE_add_cert(
                        x509_store,
                        oscrypto_cert.x509
                    )
                    handle_openssl_error(result)

        except (Exception):
            if ssl_ctx:
                libssl.SSL_CTX_free(ssl_ctx)
            self._ssl_ctx = None
            raise

    def __del__(self):
        if self._ssl_ctx:
            libssl.SSL_CTX_free(self._ssl_ctx)
            self._ssl_ctx = None

        if self._ssl_session:
            libssl.SSL_SESSION_free(self._ssl_session)
            self._ssl_session = None


class TLSSocket(object):
    """
    A wrapper around a socket.socket that adds TLS
    """

    _socket = None

    # An oscrypto.tls.TLSSession object
    _session = None

    # An OpenSSL SSL struct pointer
    _ssl = None

    # OpenSSL memory bios used for reading/writing data to and
    # from the socket
    _rbio = None
    _wbio = None

    # Size of _bio_write_buffer and _read_buffer
    _buffer_size = 8192

    # A buffer used to pull bytes out of the _wbio memory bio to
    # be written to the socket
    _bio_write_buffer = None

    # A buffer used to push bytes into the _rbio memory bio to
    # be decrypted by OpenSSL
    _read_buffer = None

    # Raw ciphertext from the socker that hasn't need fed to OpenSSL yet
    _raw_bytes = None

    # Plaintext that has been decrypted, but not asked for yet
    _decrypted_bytes = None

    _hostname = None

    _certificate = None
    _intermediates = None

    _protocol = None
    _cipher_suite = None
    _compression = None
    _session_id = None
    _session_ticket = None

    # If we explicitly asked for the connection to be closed
    _local_closed = False

    _gracefully_closed = False

    @classmethod
    def wrap(cls, socket, hostname, session=None):
        """
        Takes an existing socket and adds TLS

        :param socket:
            A socket.socket object to wrap with TLS

        :param hostname:
            A unicode string of the hostname or IP the socket is connected to

        :param session:
            An existing TLSSession object to allow for session reuse, specific
            protocol or manual certificate validation

        :raises:
            ValueError - when any of the parameters contain an invalid value
            TypeError - when any of the parameters are of the wrong type
            OSError - when an error is returned by the OS crypto library
        """

        if not isinstance(socket, socket_.socket):
            raise TypeError(pretty_message(
                '''
                socket must be an instance of socket.socket, not %s
                ''',
                type_name(socket)
            ))

        if not isinstance(hostname, str_cls):
            raise TypeError(pretty_message(
                '''
                hostname must be a unicode string, not %s
                ''',
                type_name(hostname)
            ))

        if session is not None and not isinstance(session, TLSSession):
            raise TypeError(pretty_message(
                '''
                session must be an instance of oscrypto.tls.TLSSession, not %s
                ''',
                type_name(session)
            ))

        new_socket = cls(None, None, session=session)
        new_socket._socket = socket
        new_socket._hostname = hostname
        new_socket._handshake()

        return new_socket

    def __init__(self, address, port, timeout=10, session=None):
        """
        :param address:
            A unicode string of the domain name or IP address to connect to

        :param port:
            An integer of the port number to connect to

        :param timeout:
            An integer timeout to use for the socket

        :param session:
            An oscrypto.tls.TLSSession object to allow for session reuse and
            controlling the protocols and validation performed
        """

        self._raw_bytes = b''
        self._decrypted_bytes = b''

        if address is None and port is None:
            self._socket = None

        else:
            if not isinstance(address, str_cls):
                raise TypeError(pretty_message(
                    '''
                    address must be a unicode string, not %s
                    ''',
                    type_name(address)
                ))

            if not isinstance(port, int_types):
                raise TypeError(pretty_message(
                    '''
                    port must be an integer, not %s
                    ''',
                    type_name(port)
                ))

            if timeout is not None and not isinstance(timeout, numbers.Number):
                raise TypeError(pretty_message(
                    '''
                    timeout must be a number, not %s
                    ''',
                    type_name(timeout)
                ))

            self._socket = socket_.create_connection((address, port), timeout)
            self._socket.settimeout(timeout)

        if session is None:
            session = TLSSession()

        elif not isinstance(session, TLSSession):
            raise TypeError(pretty_message(
                '''
                session must be an instance of oscrypto.tls.TLSSession, not %s
                ''',
                type_name(session)
            ))

        self._session = session

        if self._socket:
            self._hostname = address
            self._handshake()

    def _handshake(self):
        """
        Perform an initial TLS handshake
        """

        self._ssl = None
        self._rbio = None
        self._wbio = None

        try:
            self._ssl = libssl.SSL_new(self._session._ssl_ctx)
            if is_null(self._ssl):
                self._ssl = None
                handle_openssl_error(0)

            mem_bio = libssl.BIO_s_mem()

            self._rbio = libssl.BIO_new(mem_bio)
            if is_null(self._rbio):
                handle_openssl_error(0)

            self._wbio = libssl.BIO_new(mem_bio)
            if is_null(self._wbio):
                handle_openssl_error(0)

            libssl.SSL_set_bio(self._ssl, self._rbio, self._wbio)

            utf8_domain = self._hostname.encode('utf-8')
            libssl.SSL_ctrl(
                self._ssl,
                LibsslConst.SSL_CTRL_SET_TLSEXT_HOSTNAME,
                LibsslConst.TLSEXT_NAMETYPE_host_name,
                utf8_domain
            )

            libssl.SSL_set_connect_state(self._ssl)

            if self._session._ssl_session:
                libssl.SSL_set_session(self._ssl, self._session._ssl_session)

            self._bio_write_buffer = buffer_from_bytes(self._buffer_size)
            self._read_buffer = buffer_from_bytes(self._buffer_size)

            handshake_server_bytes = b''
            handshake_client_bytes = b''

            while True:
                result = libssl.SSL_do_handshake(self._ssl)
                handshake_client_bytes += self._raw_write()

                if result == 1:
                    break

                error = libssl.SSL_get_error(self._ssl, result)
                if error == LibsslConst.SSL_ERROR_WANT_READ:
                    chunk = self._raw_read()
                    if chunk == b'':
                        if handshake_server_bytes == b'':
                            raise_disconnection()
                        if detect_client_auth_request(handshake_server_bytes):
                            raise_client_auth()
                        raise_protocol_error(handshake_server_bytes)
                    handshake_server_bytes += chunk

                elif error == LibsslConst.SSL_ERROR_WANT_WRITE:
                    handshake_client_bytes += self._raw_write()

                elif error == LibsslConst.SSL_ERROR_ZERO_RETURN:
                    self._gracefully_closed = True
                    self._shutdown(False)
                    self._raise_closed()

                else:
                    info = peek_openssl_error()

                    dh_key_info_1 = (
                        LibsslConst.ERR_LIB_SSL,
                        LibsslConst.SSL_F_SSL3_CHECK_CERT_AND_ALGORITHM,
                        LibsslConst.SSL_R_DH_KEY_TOO_SMALL
                    )
                    dh_key_info_1 = _homogenize_openssl3_error(dh_key_info_1)

                    dh_key_info_2 = (
                        LibsslConst.ERR_LIB_SSL,
                        LibsslConst.SSL_F_TLS_PROCESS_SKE_DHE,
                        LibsslConst.SSL_R_DH_KEY_TOO_SMALL
                    )
                    dh_key_info_2 = _homogenize_openssl3_error(dh_key_info_2)

                    dh_key_info_3 = (
                        LibsslConst.ERR_LIB_SSL,
                        LibsslConst.SSL_F_SSL3_GET_KEY_EXCHANGE,
                        LibsslConst.SSL_R_BAD_DH_P_LENGTH
                    )
                    dh_key_info_3 = _homogenize_openssl3_error(dh_key_info_3)

                    if info == dh_key_info_1 or info == dh_key_info_2 or info == dh_key_info_3:
                        raise_dh_params()

                    if libcrypto_version_info < (1, 1):
                        unknown_protocol_info = (
                            LibsslConst.ERR_LIB_SSL,
                            LibsslConst.SSL_F_SSL23_GET_SERVER_HELLO,
                            LibsslConst.SSL_R_UNKNOWN_PROTOCOL
                        )
                    else:
                        unknown_protocol_info = (
                            LibsslConst.ERR_LIB_SSL,
                            LibsslConst.SSL_F_SSL3_GET_RECORD,
                            LibsslConst.SSL_R_WRONG_VERSION_NUMBER
                        )
                        unknown_protocol_info = _homogenize_openssl3_error(unknown_protocol_info)

                    if info == unknown_protocol_info:
                        raise_protocol_error(handshake_server_bytes)

                    tls_version_info_error = (
                        LibsslConst.ERR_LIB_SSL,
                        LibsslConst.SSL_F_SSL23_GET_SERVER_HELLO,
                        LibsslConst.SSL_R_TLSV1_ALERT_PROTOCOL_VERSION
                    )
                    tls_version_info_error = _homogenize_openssl3_error(tls_version_info_error)
                    if info == tls_version_info_error:
                        raise_protocol_version()

                    handshake_error_info = (
                        LibsslConst.ERR_LIB_SSL,
                        LibsslConst.SSL_F_SSL23_GET_SERVER_HELLO,
                        LibsslConst.SSL_R_SSLV3_ALERT_HANDSHAKE_FAILURE
                    )
                    # OpenSSL 3.0 no longer has func codes, so this can be confused
                    # with the following handler which needs to check for client auth
                    if libcrypto_version_info < (3, ) and info == handshake_error_info:
                        raise_handshake()

                    handshake_failure_info = (
                        LibsslConst.ERR_LIB_SSL,
                        LibsslConst.SSL_F_SSL3_READ_BYTES,
                        LibsslConst.SSL_R_SSLV3_ALERT_HANDSHAKE_FAILURE
                    )
                    handshake_failure_info = _homogenize_openssl3_error(handshake_failure_info)
                    if info == handshake_failure_info:
                        saw_client_auth = False
                        for record_type, _, record_data in parse_tls_records(handshake_server_bytes):
                            if record_type != b'\x16':
                                continue
                            for message_type, message_data in parse_handshake_messages(record_data):
                                if message_type == b'\x0d':
                                    saw_client_auth = True
                                    break
                        if saw_client_auth:
                            raise_client_auth()
                        raise_handshake()

                    if libcrypto_version_info < (1, 1):
                        cert_verify_failed_info = (
                            LibsslConst.ERR_LIB_SSL,
                            LibsslConst.SSL_F_SSL3_GET_SERVER_CERTIFICATE,
                            LibsslConst.SSL_R_CERTIFICATE_VERIFY_FAILED
                        )
                    else:
                        cert_verify_failed_info = (
                            LibsslConst.ERR_LIB_SSL,
                            LibsslConst.SSL_F_TLS_PROCESS_SERVER_CERTIFICATE,
                            LibsslConst.SSL_R_CERTIFICATE_VERIFY_FAILED
                        )
                        cert_verify_failed_info = _homogenize_openssl3_error(cert_verify_failed_info)

                    # It would appear that some versions of OpenSSL (such as on Fedora 30)
                    # don't even have the MD5 digest algorithm included any longer? To
                    # give a more useful error message we handle this specifically.
                    unknown_hash_algo_info = (
                        LibsslConst.ERR_LIB_ASN1,
                        LibsslConst.ASN1_F_ASN1_ITEM_VERIFY,
                        LibsslConst.ASN1_R_UNKNOWN_MESSAGE_DIGEST_ALGORITHM
                    )
                    unknown_hash_algo_info = _homogenize_openssl3_error(unknown_hash_algo_info)

                    if info == unknown_hash_algo_info:
                        chain = extract_chain(handshake_server_bytes)
                        if chain:
                            cert = chain[0]
                            oscrypto_cert = load_certificate(cert)
                            if oscrypto_cert.asn1.hash_algo in set(['md5', 'md2']):
                                raise_weak_signature(oscrypto_cert)

                    if info == cert_verify_failed_info:
                        verify_result = libssl.SSL_get_verify_result(self._ssl)
                        chain = extract_chain(handshake_server_bytes)

                        self_signed = False
                        time_invalid = False
                        no_issuer = False
                        cert = None
                        oscrypto_cert = None

                        if chain:
                            cert = chain[0]
                            oscrypto_cert = load_certificate(cert)
                            self_signed = oscrypto_cert.self_signed

                            issuer_error_codes = set([
                                LibsslConst.X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT,
                                LibsslConst.X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN,
                                LibsslConst.X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY
                            ])
                            if verify_result in issuer_error_codes:
                                no_issuer = not self_signed

                            time_error_codes = set([
                                LibsslConst.X509_V_ERR_CERT_HAS_EXPIRED,
                                LibsslConst.X509_V_ERR_CERT_NOT_YET_VALID
                            ])
                            time_invalid = verify_result in time_error_codes

                        if time_invalid:
                            raise_expired_not_yet_valid(cert)
                        if no_issuer:
                            raise_no_issuer(cert)
                        if self_signed:
                            raise_self_signed(cert)
                        if oscrypto_cert and oscrypto_cert.asn1.hash_algo in set(['md5', 'md2']):
                            raise_weak_signature(oscrypto_cert)
                        raise_verification(cert)

                    handle_openssl_error(0, TLSError)

            session_info = parse_session_info(
                handshake_server_bytes,
                handshake_client_bytes
            )
            self._protocol = session_info['protocol']
            self._cipher_suite = session_info['cipher_suite']
            self._compression = session_info['compression']
            self._session_id = session_info['session_id']
            self._session_ticket = session_info['session_ticket']

            if self._cipher_suite.find('_DHE_') != -1:
                dh_params_length = get_dh_params_length(handshake_server_bytes)
                if dh_params_length < 1024:
                    self.close()
                    raise_dh_params()

            # When saving the session for future requests, we use
            # SSL_get1_session() variant to increase the reference count. This
            # prevents the session from being freed when one connection closes
            # before another is opened. However, since we increase the ref
            # count, we also have to explicitly free any previous session.
            if self._session_id == 'new' or self._session_ticket == 'new':
                if self._session._ssl_session:
                    libssl.SSL_SESSION_free(self._session._ssl_session)
                self._session._ssl_session = libssl.SSL_get1_session(self._ssl)

            if not self._session._manual_validation:
                if self.certificate.hash_algo in set(['md5', 'md2']):
                    raise_weak_signature(self.certificate)

                # OpenSSL does not do hostname or IP address checking in the end
                # entity certificate, so we must perform that check
                if not self.certificate.is_valid_domain_ip(self._hostname):
                    raise_hostname(self.certificate, self._hostname)

        except (OSError, socket_.error):
            if self._ssl:
                libssl.SSL_free(self._ssl)
                self._ssl = None
                self._rbio = None
                self._wbio = None
            # The BIOs are freed by SSL_free(), so we only need to free
            # them if for some reason SSL_free() was not called
            else:
                if self._rbio:
                    libssl.BIO_free(self._rbio)
                    self._rbio = None
                if self._wbio:
                    libssl.BIO_free(self._wbio)
                    self._wbio = None
            self.close()

            raise

    def _raw_read(self):
        """
        Reads data from the socket and writes it to the memory bio
        used by libssl to decrypt the data. Returns the unencrypted
        data for the purpose of debugging handshakes.

        :return:
            A byte string of ciphertext from the socket. Used for
            debugging the handshake only.
        """

        data = self._raw_bytes
        try:
            data += self._socket.recv(8192)
        except (socket_.error):
            pass
        output = data
        written = libssl.BIO_write(self._rbio, data, len(data))
        self._raw_bytes = data[written:]
        return output

    def _raw_write(self):
        """
        Takes ciphertext from the memory bio and writes it to the
        socket.

        :return:
            A byte string of ciphertext going to the socket. Used
            for debugging the handshake only.
        """

        data_available = libssl.BIO_ctrl_pending(self._wbio)
        if data_available == 0:
            return b''
        to_read = min(self._buffer_size, data_available)
        read = libssl.BIO_read(self._wbio, self._bio_write_buffer, to_read)
        to_write = bytes_from_buffer(self._bio_write_buffer, read)
        output = to_write
        while len(to_write):
            raise_disconnect = False
            try:
                sent = self._socket.send(to_write)
            except (socket_.error) as e:
                # Handle ECONNRESET and EPIPE
                if e.errno == 104 or e.errno == 32:
                    raise_disconnect = True
                # Handle EPROTOTYPE. Newer versions of macOS will return this
                # if we try to call send() while the socket is being torn down
                elif sys.platform == 'darwin' and e.errno == 41:
                    raise_disconnect = True
                else:
                    raise

            if raise_disconnect:
                raise_disconnection()
            to_write = to_write[sent:]
            if len(to_write):
                self.select_write()
        return output

    def read(self, max_length):
        """
        Reads data from the TLS-wrapped socket

        :param max_length:
            The number of bytes to read - output may be less than this

        :raises:
            socket.socket - when a non-TLS socket error occurs
            oscrypto.errors.TLSError - when a TLS-related error occurs
            ValueError - when any of the parameters contain an invalid value
            TypeError - when any of the parameters are of the wrong type
            OSError - when an error is returned by the OS crypto library

        :return:
            A byte string of the data read
        """

        if not isinstance(max_length, int_types):
            raise TypeError(pretty_message(
                '''
                max_length must be an integer, not %s
                ''',
                type_name(max_length)
            ))

        buffered_length = len(self._decrypted_bytes)

        # If we already have enough buffered data, just use that
        if buffered_length >= max_length:
            output = self._decrypted_bytes[0:max_length]
            self._decrypted_bytes = self._decrypted_bytes[max_length:]
            return output

        if self._ssl is None:
            self._raise_closed()

        # Don't block if we have buffered data available, since it is ok to
        # return less than the max_length
        if buffered_length > 0 and not self.select_read(0):
            output = self._decrypted_bytes
            self._decrypted_bytes = b''
            return output

        # Only read enough to get the requested amount when
        # combined with buffered data
        to_read = min(self._buffer_size, max_length - buffered_length)

        output = self._decrypted_bytes

        # The SSL_read() loop handles renegotiations, so we need to handle
        # requests for both reads and writes
        again = True
        while again:
            again = False
            result = libssl.SSL_read(self._ssl, self._read_buffer, to_read)
            self._raw_write()
            if result <= 0:

                error = libssl.SSL_get_error(self._ssl, result)
                if error == LibsslConst.SSL_ERROR_WANT_READ:
                    if self._raw_read() != b'':
                        again = True
                        continue
                    raise_disconnection()

                elif error == LibsslConst.SSL_ERROR_WANT_WRITE:
                    self._raw_write()
                    again = True
                    continue

                elif error == LibsslConst.SSL_ERROR_ZERO_RETURN:
                    self._gracefully_closed = True
                    self._shutdown(False)
                    break

                else:
                    handle_openssl_error(0, TLSError)

            output += bytes_from_buffer(self._read_buffer, result)

        if self._gracefully_closed and len(output) == 0:
            self._raise_closed()

        self._decrypted_bytes = output[max_length:]
        return output[0:max_length]

    def select_read(self, timeout=None):
        """
        Blocks until the socket is ready to be read from, or the timeout is hit

        :param timeout:
            A float - the period of time to wait for data to be read. None for
            no time limit.

        :return:
            A boolean - if data is ready to be read. Will only be False if
            timeout is not None.
        """

        # If we have buffered data, we consider a read possible
        if len(self._decrypted_bytes) > 0:
            return True

        read_ready, _, _ = select.select([self._socket], [], [], timeout)
        return len(read_ready) > 0

    def read_until(self, marker):
        """
        Reads data from the socket until a marker is found. Data read includes
        the marker.

        :param marker:
            A byte string or regex object from re.compile(). Used to determine
            when to stop reading. Regex objects are more inefficient since
            they must scan the entire byte string of read data each time data
            is read off the socket.

        :return:
            A byte string of the data read, including the marker
        """

        if not isinstance(marker, byte_cls) and not isinstance(marker, Pattern):
            raise TypeError(pretty_message(
                '''
                marker must be a byte string or compiled regex object, not %s
                ''',
                type_name(marker)
            ))

        output = b''

        is_regex = isinstance(marker, Pattern)

        while True:
            if len(self._decrypted_bytes) > 0:
                chunk = self._decrypted_bytes
                self._decrypted_bytes = b''
            else:
                if self._ssl is None:
                    self._raise_closed()
                to_read = libssl.SSL_pending(self._ssl) or 8192
                chunk = self.read(to_read)

            offset = len(output)
            output += chunk

            if is_regex:
                match = marker.search(output)
                if match is not None:
                    end = match.end()
                    break
            else:
                # If the marker was not found last time, we have to start
                # at a position where the marker would have its final char
                # in the newly read chunk
                start = max(0, offset - len(marker) - 1)
                match = output.find(marker, start)
                if match != -1:
                    end = match + len(marker)
                    break

        self._decrypted_bytes = output[end:] + self._decrypted_bytes
        return output[0:end]

    def read_line(self):
        r"""
        Reads a line from the socket, including the line ending of "\r\n", "\r",
        or "\n"

        :return:
            A byte string of the next line from the socket
        """

        return self.read_until(_line_regex)

    def read_exactly(self, num_bytes):
        """
        Reads exactly the specified number of bytes from the socket

        :param num_bytes:
            An integer - the exact number of bytes to read

        :return:
            A byte string of the data that was read
        """

        output = b''
        remaining = num_bytes
        while remaining > 0:
            output += self.read(remaining)
            remaining = num_bytes - len(output)

        return output

    def write(self, data):
        """
        Writes data to the TLS-wrapped socket

        :param data:
            A byte string to write to the socket

        :raises:
            socket.socket - when a non-TLS socket error occurs
            oscrypto.errors.TLSError - when a TLS-related error occurs
            ValueError - when any of the parameters contain an invalid value
            TypeError - when any of the parameters are of the wrong type
            OSError - when an error is returned by the OS crypto library
        """

        data_len = len(data)
        while data_len:
            if self._ssl is None:
                self._raise_closed()
            result = libssl.SSL_write(self._ssl, data, data_len)
            self._raw_write()
            if result <= 0:

                error = libssl.SSL_get_error(self._ssl, result)
                if error == LibsslConst.SSL_ERROR_WANT_READ:
                    if self._raw_read() != b'':
                        continue
                    raise_disconnection()

                elif error == LibsslConst.SSL_ERROR_WANT_WRITE:
                    self._raw_write()
                    continue

                elif error == LibsslConst.SSL_ERROR_ZERO_RETURN:
                    self._gracefully_closed = True
                    self._shutdown(False)
                    self._raise_closed()

                else:
                    handle_openssl_error(0, TLSError)

            data = data[result:]
            data_len = len(data)

    def select_write(self, timeout=None):
        """
        Blocks until the socket is ready to be written to, or the timeout is hit

        :param timeout:
            A float - the period of time to wait for the socket to be ready to
            written to. None for no time limit.

        :return:
            A boolean - if the socket is ready for writing. Will only be False
            if timeout is not None.
        """

        _, write_ready, _ = select.select([], [self._socket], [], timeout)
        return len(write_ready) > 0

    def _shutdown(self, manual):
        """
        Shuts down the TLS session and then shuts down the underlying socket

        :param manual:
            A boolean if the connection was manually shutdown
        """

        if self._ssl is None:
            return

        while True:
            result = libssl.SSL_shutdown(self._ssl)

            # Don't be noisy if the socket is already closed
            try:
                self._raw_write()
            except (TLSDisconnectError):
                pass

            if result >= 0:
                break
            if result < 0:
                error = libssl.SSL_get_error(self._ssl, result)
                if error == LibsslConst.SSL_ERROR_WANT_READ:
                    if self._raw_read() != b'':
                        continue
                    else:
                        break

                elif error == LibsslConst.SSL_ERROR_WANT_WRITE:
                    self._raw_write()
                    continue

                else:
                    handle_openssl_error(0, TLSError)

        if manual:
            self._local_closed = True

        libssl.SSL_free(self._ssl)
        self._ssl = None
        # BIOs are freed by SSL_free()
        self._rbio = None
        self._wbio = None

        try:
            self._socket.shutdown(socket_.SHUT_RDWR)
        except (socket_.error):
            pass

    def shutdown(self):
        """
        Shuts down the TLS session and then shuts down the underlying socket
        """

        self._shutdown(True)

    def close(self):
        """
        Shuts down the TLS session and socket and forcibly closes it
        """

        try:
            self.shutdown()

        finally:
            if self._socket:
                try:
                    self._socket.close()
                except (socket_.error):
                    pass
                self._socket = None

    def _read_certificates(self):
        """
        Reads end-entity and intermediate certificate information from the
        TLS session
        """

        stack_pointer = libssl.SSL_get_peer_cert_chain(self._ssl)
        if is_null(stack_pointer):
            handle_openssl_error(0, TLSError)

        if libcrypto_version_info < (1, 1):
            number_certs = libssl.sk_num(stack_pointer)
        else:
            number_certs = libssl.OPENSSL_sk_num(stack_pointer)

        self._intermediates = []

        for index in range(0, number_certs):
            if libcrypto_version_info < (1, 1):
                x509_ = libssl.sk_value(stack_pointer, index)
            else:
                x509_ = libssl.OPENSSL_sk_value(stack_pointer, index)
            buffer_size = libcrypto.i2d_X509(x509_, null())
            cert_buffer = buffer_from_bytes(buffer_size)
            cert_pointer = buffer_pointer(cert_buffer)
            cert_length = libcrypto.i2d_X509(x509_, cert_pointer)
            handle_openssl_error(cert_length)
            cert_data = bytes_from_buffer(cert_buffer, cert_length)

            cert = Asn1Certificate.load(cert_data)

            if index == 0:
                self._certificate = cert
            else:
                self._intermediates.append(cert)

    def _raise_closed(self):
        """
        Raises an exception describing if the local or remote end closed the
        connection
        """

        if self._local_closed:
            raise TLSDisconnectError('The connection was already closed')
        elif self._gracefully_closed:
            raise TLSGracefulDisconnectError('The remote end closed the connection')
        else:
            raise TLSDisconnectError('The connection was closed')

    @property
    def certificate(self):
        """
        An asn1crypto.x509.Certificate object of the end-entity certificate
        presented by the server
        """

        if self._ssl is None:
            self._raise_closed()

        if self._certificate is None:
            self._read_certificates()

        return self._certificate

    @property
    def intermediates(self):
        """
        A list of asn1crypto.x509.Certificate objects that were presented as
        intermediates by the server
        """

        if self._ssl is None:
            self._raise_closed()

        if self._certificate is None:
            self._read_certificates()

        return self._intermediates

    @property
    def cipher_suite(self):
        """
        A unicode string of the IANA cipher suite name of the negotiated
        cipher suite
        """

        return self._cipher_suite

    @property
    def protocol(self):
        """
        A unicode string of: "TLSv1.2", "TLSv1.1", "TLSv1", "SSLv3"
        """

        return self._protocol

    @property
    def compression(self):
        """
        A boolean if compression is enabled
        """

        return self._compression

    @property
    def session_id(self):
        """
        A unicode string of "new" or "reused" or None for no ticket
        """

        return self._session_id

    @property
    def session_ticket(self):
        """
        A unicode string of "new" or "reused" or None for no ticket
        """

        return self._session_ticket

    @property
    def session(self):
        """
        The oscrypto.tls.TLSSession object used for this connection
        """

        return self._session

    @property
    def hostname(self):
        """
        A unicode string of the TLS server domain name or IP address
        """

        return self._hostname

    @property
    def port(self):
        """
        An integer of the port number the socket is connected to
        """

        return self.socket.getpeername()[1]

    @property
    def socket(self):
        """
        The underlying socket.socket connection
        """

        if self._ssl is None:
            self._raise_closed()

        return self._socket

    def __del__(self):
        self.close()



# coding: utf-8

"""
Function to fill ensure integers converted to a byte string are a specific
width. Exports the following items:

 - fill_width()
"""

from __future__ import unicode_literals, division, absolute_import, print_function


__all__ = [
    'fill_width',
]


def fill_width(bytes_, width):
    """
    Ensure a byte string representing a positive integer is a specific width
    (in bytes)

    :param bytes_:
        The integer byte string

    :param width:
        The desired width as an integer

    :return:
        A byte string of the width specified
    """

    while len(bytes_) < width:
        bytes_ = b'\x00' + bytes_
    return bytes_



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

from .. import ffi

# Initialize OpenSSL
from ._libcrypto import libcrypto_version_info

if ffi() == 'cffi':
    from ._libssl_cffi import libssl
else:
    from ._libssl_ctypes import libssl


__all__ = [
    'libssl',
    'LibsslConst',
]


if libcrypto_version_info < (1, 1):
    libssl.SSL_library_init()
# Enables SHA2 algorithms on 0.9.8n and older
if libcrypto_version_info < (1, 0):
    libssl.OPENSSL_add_all_algorithms_noconf()


class LibsslConst():
    ERR_LIB_ASN1 = 13
    ERR_LIB_SSL = 20

    SSL_CTRL_OPTIONS = 32
    SSL_CTRL_SET_SESS_CACHE_MODE = 44

    SSL_VERIFY_NONE = 0
    SSL_VERIFY_PEER = 1

    SSL_ST_OK = 3

    SSL_ERROR_WANT_READ = 2
    SSL_ERROR_WANT_WRITE = 3
    SSL_ERROR_ZERO_RETURN = 6

    SSL_OP_NO_SSLv2 = 0x01000000
    SSL_OP_NO_SSLv3 = 0x02000000
    SSL_OP_NO_TLSv1 = 0x04000000
    SSL_OP_NO_TLSv1_2 = 0x08000000
    SSL_OP_NO_TLSv1_1 = 0x10000000

    SSL_SESS_CACHE_CLIENT = 0x0001

    SSL_R_NO_SHARED_CIPHER = 193

    SSL_F_SSL3_CHECK_CERT_AND_ALGORITHM = 130
    SSL_F_SSL3_GET_KEY_EXCHANGE = 141
    SSL_F_SSL3_GET_SERVER_CERTIFICATE = 144
    SSL_R_BAD_DH_P_LENGTH = 110
    SSL_R_CERTIFICATE_VERIFY_FAILED = 134
    SSL_R_UNKNOWN_PROTOCOL = 252
    SSL_R_DH_KEY_TOO_SMALL = 372

    # OpenSSL 1.1.0
    SSL_F_TLS_PROCESS_SKE_DHE = 419
    SSL_F_SSL3_GET_RECORD = 143
    SSL_R_WRONG_VERSION_NUMBER = 267
    SSL_F_TLS_PROCESS_SERVER_CERTIFICATE = 367

    # OpenSSL < 1.1.0
    SSL_F_SSL23_GET_SERVER_HELLO = 119
    SSL_F_SSL3_READ_BYTES = 148
    SSL_R_SSLV3_ALERT_HANDSHAKE_FAILURE = 1040
    SSL_R_TLSV1_ALERT_PROTOCOL_VERSION = 1070

    SSL_CTRL_SET_TLSEXT_HOSTNAME = 55
    TLSEXT_NAMETYPE_host_name = 0

    X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY = 20
    X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN = 19
    X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT = 18

    X509_V_ERR_CERT_NOT_YET_VALID = 9
    X509_V_ERR_CERT_HAS_EXPIRED = 10

    ASN1_F_ASN1_ITEM_VERIFY = 197
    ASN1_R_UNKNOWN_MESSAGE_DIGEST_ALGORITHM = 161


if libcrypto_version_info >= (1, 1, 0):
    LibsslConst.SSL_R_DH_KEY_TOO_SMALL = 394



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

from .._ffi import register_ffi
from .._types import str_cls
from ..errors import LibraryNotFoundError

import cffi


__all__ = [
    'advapi32',
    'get_error',
]


ffi = cffi.FFI()
if cffi.__version_info__ >= (0, 9):
    ffi.set_unicode(True)
ffi.cdef("""
    typedef HANDLE HCRYPTPROV;
    typedef HANDLE HCRYPTKEY;
    typedef HANDLE HCRYPTHASH;
    typedef unsigned int ALG_ID;

    typedef struct _CRYPTOAPI_BLOB {
        DWORD cbData;
        BYTE  *pbData;
    } CRYPT_INTEGER_BLOB, CRYPT_OBJID_BLOB, CRYPT_DER_BLOB, CRYPT_ATTR_BLOB;

    typedef struct _CRYPT_ALGORITHM_IDENTIFIER {
        LPSTR            pszObjId;
        CRYPT_OBJID_BLOB Parameters;
    } CRYPT_ALGORITHM_IDENTIFIER;

    typedef struct _CRYPT_BIT_BLOB {
        DWORD cbData;
        BYTE  *pbData;
        DWORD cUnusedBits;
    } CRYPT_BIT_BLOB;

    typedef struct _CERT_PUBLIC_KEY_INFO {
        CRYPT_ALGORITHM_IDENTIFIER Algorithm;
        CRYPT_BIT_BLOB             PublicKey;
    } CERT_PUBLIC_KEY_INFO;

    typedef struct _CRYPT_ATTRIBUTE {
        LPSTR           pszObjId;
        DWORD           cValue;
        CRYPT_ATTR_BLOB *rgValue;
    } CRYPT_ATTRIBUTE;

    typedef struct _CRYPT_ATTRIBUTES {
        DWORD           cAttr;
        CRYPT_ATTRIBUTE *rgAttr;
    } CRYPT_ATTRIBUTES;

    typedef struct _CRYPT_PRIVATE_KEY_INFO {
        DWORD                      Version;
        CRYPT_ALGORITHM_IDENTIFIER Algorithm;
        CRYPT_DER_BLOB             PrivateKey;
        CRYPT_ATTRIBUTES           *pAttributes;
    } CRYPT_PRIVATE_KEY_INFO;

    typedef struct _PUBLICKEYSTRUC {
        BYTE   bType;
        BYTE   bVersion;
        WORD   reserved;
        ALG_ID aiKeyAlg;
    } BLOBHEADER, PUBLICKEYSTRUC;

    typedef struct _DSSPUBKEY {
        DWORD magic;
        DWORD bitlen;
    } DSSPUBKEY;

    typedef struct _DSSBLOBHEADER {
        PUBLICKEYSTRUC  publickeystruc;
        DSSPUBKEY dsspubkey;
    } DSSBLOBHEADER;

    typedef struct _RSAPUBKEY {
        DWORD magic;
        DWORD bitlen;
        DWORD pubexp;
    } RSAPUBKEY;

    typedef struct _RSABLOBHEADER {
        PUBLICKEYSTRUC  publickeystruc;
        RSAPUBKEY rsapubkey;
    } RSABLOBHEADER;

    typedef struct _PLAINTEXTKEYBLOB {
        BLOBHEADER hdr;
        DWORD      dwKeySize;
        // rgbKeyData omitted since it is a flexible array member
    } PLAINTEXTKEYBLOB;

    typedef struct _DSSSEED {
        DWORD counter;
        BYTE  seed[20];
    } DSSSEED;

    BOOL CryptAcquireContextW(HCRYPTPROV *phProv, LPCWSTR pszContainer, LPCWSTR pszProvider,
                DWORD dwProvType, DWORD dwFlags);
    BOOL CryptReleaseContext(HCRYPTPROV hProv, DWORD dwFlags);

    BOOL CryptImportKey(HCRYPTPROV hProv, BYTE *pbData, DWORD dwDataLen,
                HCRYPTKEY hPubKey, DWORD dwFlags, HCRYPTKEY *phKey);
    BOOL CryptGenKey(HCRYPTPROV hProv, ALG_ID Algid, DWORD dwFlags, HCRYPTKEY *phKey);
    BOOL CryptGetKeyParam(HCRYPTKEY hKey, DWORD dwParam, BYTE *pbData, DWORD *pdwDataLen, DWORD dwFlags);
    BOOL CryptSetKeyParam(HCRYPTKEY hKey, DWORD dwParam, void *pbData, DWORD dwFlags);
    BOOL CryptExportKey(HCRYPTKEY hKey, HCRYPTKEY hExpKey, DWORD dwBlobType,
                DWORD dwFlags, BYTE *pbData, DWORD *pdwDataLen);
    BOOL CryptDestroyKey(HCRYPTKEY hKey);

    BOOL CryptCreateHash(HCRYPTPROV hProv, ALG_ID Algid, HCRYPTKEY hKey,
                DWORD dwFlags, HCRYPTHASH *phHash);
    BOOL CryptHashData(HCRYPTHASH hHash, BYTE *pbData, DWORD dwDataLen, DWORD dwFlags);
    BOOL CryptSetHashParam(HCRYPTHASH hHash, DWORD dwParam, BYTE *pbData, DWORD dwFlags);
    BOOL CryptSignHashW(HCRYPTHASH hHash, DWORD dwKeySpec, LPCWSTR sDescription,
                DWORD dwFlags, BYTE *pbSignature, DWORD *pdwSigLen);
    BOOL CryptVerifySignatureW(HCRYPTHASH hHash, BYTE *pbSignature, DWORD dwSigLen,
                HCRYPTKEY hPubKey, LPCWSTR sDescription, DWORD dwFlags);
    BOOL CryptDestroyHash(HCRYPTHASH hHash);

    BOOL CryptEncrypt(HCRYPTKEY hKey, HCRYPTHASH hHash, BOOL Final, DWORD dwFlags,
                BYTE *pbData, DWORD *pdwDataLen, DWORD dwBufLen);
    BOOL CryptDecrypt(HCRYPTKEY hKey, HCRYPTHASH hHash, BOOL Final, DWORD dwFlags,
                BYTE *pbData, DWORD *pdwDataLen);
""")


try:
    advapi32 = ffi.dlopen('advapi32.dll')
    register_ffi(advapi32, ffi)

except (OSError) as e:
    if str_cls(e).find('cannot load library') != -1:
        raise LibraryNotFoundError('advapi32.dll could not be found')
    raise


def get_error():
    return ffi.getwinerror()



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import math

from .._errors import pretty_message
from .._ffi import new, null, is_null, buffer_from_bytes, bytes_from_buffer, deref
from ._libcrypto import libcrypto, libcrypto_legacy_support, LibcryptoConst, handle_openssl_error
from ..util import rand_bytes
from .._types import type_name, byte_cls


__all__ = [
    'aes_cbc_no_padding_decrypt',
    'aes_cbc_no_padding_encrypt',
    'aes_cbc_pkcs7_decrypt',
    'aes_cbc_pkcs7_encrypt',
    'des_cbc_pkcs5_decrypt',
    'des_cbc_pkcs5_encrypt',
    'rc2_cbc_pkcs5_decrypt',
    'rc2_cbc_pkcs5_encrypt',
    'rc4_decrypt',
    'rc4_encrypt',
    'tripledes_cbc_pkcs5_decrypt',
    'tripledes_cbc_pkcs5_encrypt',
]


def aes_cbc_no_padding_encrypt(key, data, iv):
    """
    Encrypts plaintext using AES in CBC mode with a 128, 192 or 256 bit key and
    no padding. This means the ciphertext must be an exact multiple of 16 bytes
    long.

    :param key:
        The encryption key - a byte string either 16, 24 or 32 bytes long

    :param data:
        The plaintext - a byte string

    :param iv:
        The initialization vector - either a byte string 16-bytes long or None
        to generate an IV

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by OpenSSL

    :return:
        A tuple of two byte strings (iv, ciphertext)
    """

    cipher = _calculate_aes_cipher(key)

    if not iv:
        iv = rand_bytes(16)
    elif len(iv) != 16:
        raise ValueError(pretty_message(
            '''
            iv must be 16 bytes long - is %s
            ''',
            len(iv)
        ))

    if len(data) % 16 != 0:
        raise ValueError(pretty_message(
            '''
            data must be a multiple of 16 bytes long - is %s
            ''',
            len(data)
        ))

    return (iv, _encrypt(cipher, key, data, iv, False))


def aes_cbc_no_padding_decrypt(key, data, iv):
    """
    Decrypts AES ciphertext in CBC mode using a 128, 192 or 256 bit key and no
    padding.

    :param key:
        The encryption key - a byte string either 16, 24 or 32 bytes long

    :param data:
        The ciphertext - a byte string

    :param iv:
        The initialization vector - a byte string 16-bytes long

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by OpenSSL

    :return:
        A byte string of the plaintext
    """

    cipher = _calculate_aes_cipher(key)

    if len(iv) != 16:
        raise ValueError(pretty_message(
            '''
            iv must be 16 bytes long - is %s
            ''',
            len(iv)
        ))

    return _decrypt(cipher, key, data, iv, False)


def aes_cbc_pkcs7_encrypt(key, data, iv):
    """
    Encrypts plaintext using AES in CBC mode with a 128, 192 or 256 bit key and
    PKCS#7 padding.

    :param key:
        The encryption key - a byte string either 16, 24 or 32 bytes long

    :param data:
        The plaintext - a byte string

    :param iv:
        The initialization vector - either a byte string 16-bytes long or None
        to generate an IV

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by OpenSSL

    :return:
        A tuple of two byte strings (iv, ciphertext)
    """

    cipher = _calculate_aes_cipher(key)

    if not iv:
        iv = rand_bytes(16)
    elif len(iv) != 16:
        raise ValueError(pretty_message(
            '''
            iv must be 16 bytes long - is %s
            ''',
            len(iv)
        ))

    return (iv, _encrypt(cipher, key, data, iv, True))


def aes_cbc_pkcs7_decrypt(key, data, iv):
    """
    Decrypts AES ciphertext in CBC mode using a 128, 192 or 256 bit key

    :param key:
        The encryption key - a byte string either 16, 24 or 32 bytes long

    :param data:
        The ciphertext - a byte string

    :param iv:
        The initialization vector - a byte string 16-bytes long

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by OpenSSL

    :return:
        A byte string of the plaintext
    """

    cipher = _calculate_aes_cipher(key)

    if len(iv) != 16:
        raise ValueError(pretty_message(
            '''
            iv must be 16 bytes long - is %s
            ''',
            len(iv)
        ))

    return _decrypt(cipher, key, data, iv, True)


def _calculate_aes_cipher(key):
    """
    Determines if the key is a valid AES 128, 192 or 256 key

    :param key:
        A byte string of the key to use

    :raises:
        ValueError - when an invalid key is provided

    :return:
        A unicode string of the AES variation - "aes128", "aes192" or "aes256"
    """

    if len(key) not in [16, 24, 32]:
        raise ValueError(pretty_message(
            '''
            key must be either 16, 24 or 32 bytes (128, 192 or 256 bits)
            long - is %s
            ''',
            len(key)
        ))

    if len(key) == 16:
        cipher = 'aes128'
    elif len(key) == 24:
        cipher = 'aes192'
    elif len(key) == 32:
        cipher = 'aes256'

    return cipher


def rc4_encrypt(key, data):
    """
    Encrypts plaintext using RC4 with a 40-128 bit key

    :param key:
        The encryption key - a byte string 5-16 bytes long

    :param data:
        The plaintext - a byte string

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by OpenSSL

    :return:
        A byte string of the ciphertext
    """

    if not libcrypto_legacy_support:
        raise EnvironmentError('OpenSSL has been compiled without RC4 support')

    if len(key) < 5 or len(key) > 16:
        raise ValueError(pretty_message(
            '''
            key must be 5 to 16 bytes (40 to 128 bits) long - is %s
            ''',
            len(key)
        ))

    return _encrypt('rc4', key, data, None, None)


def rc4_decrypt(key, data):
    """
    Decrypts RC4 ciphertext using a 40-128 bit key

    :param key:
        The encryption key - a byte string 5-16 bytes long

    :param data:
        The ciphertext - a byte string

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by OpenSSL

    :return:
        A byte string of the plaintext
    """

    if not libcrypto_legacy_support:
        raise EnvironmentError('OpenSSL has been compiled without RC4 support')

    if len(key) < 5 or len(key) > 16:
        raise ValueError(pretty_message(
            '''
            key must be 5 to 16 bytes (40 to 128 bits) long - is %s
            ''',
            len(key)
        ))

    return _decrypt('rc4', key, data, None, None)


def rc2_cbc_pkcs5_encrypt(key, data, iv):
    """
    Encrypts plaintext using RC2 in CBC mode with a 40-128 bit key and PKCS#5
    padding.

    :param key:
        The encryption key - a byte string 8 bytes long

    :param data:
        The plaintext - a byte string

    :param iv:
        The initialization vector - a byte string 8-bytes long or None
        to generate an IV

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by OpenSSL

    :return:
        A tuple of two byte strings (iv, ciphertext)
    """

    if not libcrypto_legacy_support:
        raise EnvironmentError('OpenSSL has been compiled without RC2 support')

    if len(key) < 5 or len(key) > 16:
        raise ValueError(pretty_message(
            '''
            key must be 5 to 16 bytes (40 to 128 bits) long - is %s
            ''',
            len(key)
        ))

    if not iv:
        iv = rand_bytes(8)
    elif len(iv) != 8:
        raise ValueError(pretty_message(
            '''
            iv must be 8 bytes long - is %s
            ''',
            len(iv)
        ))

    return (iv, _encrypt('rc2', key, data, iv, True))


def rc2_cbc_pkcs5_decrypt(key, data, iv):
    """
    Decrypts RC2 ciphertext ib CBC mode using a 40-128 bit key and PKCS#5
    padding.

    :param key:
        The encryption key - a byte string 8 bytes long

    :param data:
        The ciphertext - a byte string

    :param iv:
        The initialization vector - a byte string 8 bytes long

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by OpenSSL

    :return:
        A byte string of the plaintext
    """

    if not libcrypto_legacy_support:
        raise EnvironmentError('OpenSSL has been compiled without RC2 support')

    if len(key) < 5 or len(key) > 16:
        raise ValueError(pretty_message(
            '''
            key must be 5 to 16 bytes (40 to 128 bits) long - is %s
            ''',
            len(key)
        ))

    if len(iv) != 8:
        raise ValueError(pretty_message(
            '''
            iv must be 8 bytes long - is %s
            ''',
            len(iv)
        ))

    return _decrypt('rc2', key, data, iv, True)


def tripledes_cbc_pkcs5_encrypt(key, data, iv):
    """
    Encrypts plaintext using 3DES in CBC mode using either the 2 or 3 key
    variant (16 or 24 byte long key) and PKCS#5 padding.

    :param key:
        The encryption key - a byte string 16 or 24 bytes long (2 or 3 key mode)

    :param data:
        The plaintext - a byte string

    :param iv:
        The initialization vector - a byte string 8-bytes long or None
        to generate an IV

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by OpenSSL

    :return:
        A tuple of two byte strings (iv, ciphertext)
    """

    if len(key) != 16 and len(key) != 24:
        raise ValueError(pretty_message(
            '''
            key must be 16 bytes (2 key) or 24 bytes (3 key) long - %s
            ''',
            len(key)
        ))

    if not iv:
        iv = rand_bytes(8)
    elif len(iv) != 8:
        raise ValueError(pretty_message(
            '''
            iv must be 8 bytes long - %s
            ''',
            len(iv)
        ))

    cipher = 'tripledes_3key'
    # Expand 2-key to actual 24 byte byte string used by cipher
    if len(key) == 16:
        key = key + key[0:8]
        cipher = 'tripledes_2key'

    return (iv, _encrypt(cipher, key, data, iv, True))


def tripledes_cbc_pkcs5_decrypt(key, data, iv):
    """
    Decrypts 3DES ciphertext in CBC mode using either the 2 or 3 key variant
    (16 or 24 byte long key) and PKCS#5 padding.

    :param key:
        The encryption key - a byte string 16 or 24 bytes long (2 or 3 key mode)

    :param data:
        The ciphertext - a byte string

    :param iv:
        The initialization vector - a byte string 8-bytes long

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by OpenSSL

    :return:
        A byte string of the plaintext
    """

    if len(key) != 16 and len(key) != 24:
        raise ValueError(pretty_message(
            '''
            key must be 16 bytes (2 key) or 24 bytes (3 key) long - is %s
            ''',
            len(key)
        ))

    if len(iv) != 8:
        raise ValueError(pretty_message(
            '''
            iv must be 8 bytes long - is %s
            ''',
            len(iv)
        ))

    cipher = 'tripledes_3key'
    # Expand 2-key to actual 24 byte byte string used by cipher
    if len(key) == 16:
        key = key + key[0:8]
        cipher = 'tripledes_2key'

    return _decrypt(cipher, key, data, iv, True)


def des_cbc_pkcs5_encrypt(key, data, iv):
    """
    Encrypts plaintext using DES in CBC mode with a 56 bit key and PKCS#5
    padding.

    :param key:
        The encryption key - a byte string 8 bytes long (includes error correction bits)

    :param data:
        The plaintext - a byte string

    :param iv:
        The initialization vector - a byte string 8-bytes long or None
        to generate an IV

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by OpenSSL

    :return:
        A tuple of two byte strings (iv, ciphertext)
    """

    if not libcrypto_legacy_support:
        raise EnvironmentError('OpenSSL has been compiled without DES support')

    if len(key) != 8:
        raise ValueError(pretty_message(
            '''
            key must be 8 bytes (56 bits + 8 parity bits) long - is %s
            ''',
            len(key)
        ))

    if not iv:
        iv = rand_bytes(8)
    elif len(iv) != 8:
        raise ValueError(pretty_message(
            '''
            iv must be 8 bytes long - is %s
            ''',
            len(iv)
        ))

    return (iv, _encrypt('des', key, data, iv, True))


def des_cbc_pkcs5_decrypt(key, data, iv):
    """
    Decrypts DES ciphertext in CBC mode using a 56 bit key and PKCS#5 padding.

    :param key:
        The encryption key - a byte string 8 bytes long (includes error correction bits)

    :param data:
        The ciphertext - a byte string

    :param iv:
        The initialization vector - a byte string 8-bytes long

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by OpenSSL

    :return:
        A byte string of the plaintext
    """

    if not libcrypto_legacy_support:
        raise EnvironmentError('OpenSSL has been compiled without DES support')

    if len(key) != 8:
        raise ValueError(pretty_message(
            '''
            key must be 8 bytes (56 bits + 8 parity bits) long - is %s
            ''',
            len(key)
        ))

    if len(iv) != 8:
        raise ValueError(pretty_message(
            '''
            iv must be 8 bytes long - is %s
            ''',
            len(iv)
        ))

    return _decrypt('des', key, data, iv, True)


def _encrypt(cipher, key, data, iv, padding):
    """
    Encrypts plaintext

    :param cipher:
        A unicode string of "aes128", "aes192", "aes256", "des",
        "tripledes_2key", "tripledes_3key", "rc2", "rc4"

    :param key:
        The encryption key - a byte string 5-32 bytes long

    :param data:
        The plaintext - a byte string

    :param iv:
        The initialization vector - a byte string - unused for RC4

    :param padding:
        Boolean, if padding should be used - unused for RC4

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by OpenSSL

    :return:
        A byte string of the ciphertext
    """

    if not isinstance(key, byte_cls):
        raise TypeError(pretty_message(
            '''
            key must be a byte string, not %s
            ''',
            type_name(key)
        ))

    if not isinstance(data, byte_cls):
        raise TypeError(pretty_message(
            '''
            data must be a byte string, not %s
            ''',
            type_name(data)
        ))

    if cipher != 'rc4' and not isinstance(iv, byte_cls):
        raise TypeError(pretty_message(
            '''
            iv must be a byte string, not %s
            ''',
            type_name(iv)
        ))

    if cipher != 'rc4' and not padding:
        # AES in CBC mode can be allowed with no padding if
        # the data is an exact multiple of the block size
        is_aes = cipher in set(['aes128', 'aes192', 'aes256'])
        if not is_aes or (is_aes and (len(data) % 16) != 0):
            raise ValueError('padding must be specified')

    evp_cipher_ctx = None

    try:
        evp_cipher_ctx = libcrypto.EVP_CIPHER_CTX_new()
        if is_null(evp_cipher_ctx):
            handle_openssl_error(0)

        evp_cipher, buffer_size = _setup_evp_encrypt_decrypt(cipher, data)

        if iv is None:
            iv = null()

        if cipher in set(['rc2', 'rc4']):
            res = libcrypto.EVP_EncryptInit_ex(evp_cipher_ctx, evp_cipher, null(), null(), null())
            handle_openssl_error(res)
            res = libcrypto.EVP_CIPHER_CTX_set_key_length(evp_cipher_ctx, len(key))
            handle_openssl_error(res)
            if cipher == 'rc2':
                res = libcrypto.EVP_CIPHER_CTX_ctrl(
                    evp_cipher_ctx,
                    LibcryptoConst.EVP_CTRL_SET_RC2_KEY_BITS,
                    len(key) * 8,
                    null()
                )
                handle_openssl_error(res)
            evp_cipher = null()

        res = libcrypto.EVP_EncryptInit_ex(evp_cipher_ctx, evp_cipher, null(), key, iv)
        handle_openssl_error(res)

        if padding is not None:
            res = libcrypto.EVP_CIPHER_CTX_set_padding(evp_cipher_ctx, int(padding))
            handle_openssl_error(res)

        buffer = buffer_from_bytes(buffer_size)
        output_length = new(libcrypto, 'int *')

        res = libcrypto.EVP_EncryptUpdate(evp_cipher_ctx, buffer, output_length, data, len(data))
        handle_openssl_error(res)

        output = bytes_from_buffer(buffer, deref(output_length))

        res = libcrypto.EVP_EncryptFinal_ex(evp_cipher_ctx, buffer, output_length)
        handle_openssl_error(res)

        output += bytes_from_buffer(buffer, deref(output_length))

        return output

    finally:
        if evp_cipher_ctx:
            libcrypto.EVP_CIPHER_CTX_free(evp_cipher_ctx)


def _decrypt(cipher, key, data, iv, padding):
    """
    Decrypts AES/RC4/RC2/3DES/DES ciphertext

    :param cipher:
        A unicode string of "aes128", "aes192", "aes256", "des",
        "tripledes_2key", "tripledes_3key", "rc2", "rc4"

    :param key:
        The encryption key - a byte string 5-32 bytes long

    :param data:
        The ciphertext - a byte string

    :param iv:
        The initialization vector - a byte string - unused for RC4

    :param padding:
        Boolean, if padding should be used - unused for RC4

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by OpenSSL

    :return:
        A byte string of the plaintext
    """

    if not isinstance(key, byte_cls):
        raise TypeError(pretty_message(
            '''
            key must be a byte string, not %s
            ''',
            type_name(key)
        ))

    if not isinstance(data, byte_cls):
        raise TypeError(pretty_message(
            '''
            data must be a byte string, not %s
            ''',
            type_name(data)
        ))

    if cipher != 'rc4' and not isinstance(iv, byte_cls):
        raise TypeError(pretty_message(
            '''
            iv must be a byte string, not %s
            ''',
            type_name(iv)
        ))

    if cipher not in set(['rc4', 'aes128', 'aes192', 'aes256']) and not padding:
        raise ValueError('padding must be specified')

    evp_cipher_ctx = None

    try:
        evp_cipher_ctx = libcrypto.EVP_CIPHER_CTX_new()
        if is_null(evp_cipher_ctx):
            handle_openssl_error(0)

        evp_cipher, buffer_size = _setup_evp_encrypt_decrypt(cipher, data)

        if iv is None:
            iv = null()

        if cipher in set(['rc2', 'rc4']):
            res = libcrypto.EVP_DecryptInit_ex(evp_cipher_ctx, evp_cipher, null(), null(), null())
            handle_openssl_error(res)
            res = libcrypto.EVP_CIPHER_CTX_set_key_length(evp_cipher_ctx, len(key))
            handle_openssl_error(res)
            if cipher == 'rc2':
                res = libcrypto.EVP_CIPHER_CTX_ctrl(
                    evp_cipher_ctx,
                    LibcryptoConst.EVP_CTRL_SET_RC2_KEY_BITS,
                    len(key) * 8,
                    null()
                )
                handle_openssl_error(res)
            evp_cipher = null()

        res = libcrypto.EVP_DecryptInit_ex(evp_cipher_ctx, evp_cipher, null(), key, iv)
        handle_openssl_error(res)

        if padding is not None:
            res = libcrypto.EVP_CIPHER_CTX_set_padding(evp_cipher_ctx, int(padding))
            handle_openssl_error(res)

        buffer = buffer_from_bytes(buffer_size)
        output_length = new(libcrypto, 'int *')

        res = libcrypto.EVP_DecryptUpdate(evp_cipher_ctx, buffer, output_length, data, len(data))
        handle_openssl_error(res)

        output = bytes_from_buffer(buffer, deref(output_length))

        res = libcrypto.EVP_DecryptFinal_ex(evp_cipher_ctx, buffer, output_length)
        handle_openssl_error(res)

        output += bytes_from_buffer(buffer, deref(output_length))

        return output

    finally:
        if evp_cipher_ctx:
            libcrypto.EVP_CIPHER_CTX_free(evp_cipher_ctx)


def _setup_evp_encrypt_decrypt(cipher, data):
    """
    Creates an EVP_CIPHER pointer object and determines the buffer size
    necessary for the parameter specified.

    :param evp_cipher_ctx:
        An EVP_CIPHER_CTX pointer

    :param cipher:
        A unicode string of "aes128", "aes192", "aes256", "des",
        "tripledes_2key", "tripledes_3key", "rc2", "rc4"

    :param key:
        The key byte string

    :param data:
        The plaintext or ciphertext as a byte string

    :param padding:
        If padding is to be used

    :return:
        A 2-element tuple with the first element being an EVP_CIPHER pointer
        and the second being an integer that is the required buffer size
    """

    evp_cipher = {
        'aes128': libcrypto.EVP_aes_128_cbc,
        'aes192': libcrypto.EVP_aes_192_cbc,
        'aes256': libcrypto.EVP_aes_256_cbc,
        'rc2': libcrypto.EVP_rc2_cbc,
        'rc4': libcrypto.EVP_rc4,
        'des': libcrypto.EVP_des_cbc,
        'tripledes_2key': libcrypto.EVP_des_ede_cbc,
        'tripledes_3key': libcrypto.EVP_des_ede3_cbc,
    }[cipher]()

    if cipher == 'rc4':
        buffer_size = len(data)
    else:
        block_size = {
            'aes128': 16,
            'aes192': 16,
            'aes256': 16,
            'rc2': 8,
            'des': 8,
            'tripledes_2key': 8,
            'tripledes_3key': 8,
        }[cipher]
        buffer_size = block_size * int(math.ceil(len(data) / block_size))

    return (evp_cipher, buffer_size)



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import os

from .._errors import pretty_message
from .._ffi import buffer_from_bytes, bytes_from_buffer, errno, byte_string_from_buffer
from .._types import type_name, str_cls, byte_cls, int_types
from ..errors import LibraryNotFoundError
from ._common_crypto import CommonCrypto, CommonCryptoConst
from ._security import Security


__all__ = [
    'pbkdf2',
    'pkcs12_kdf',
    'rand_bytes',
]


_encoding = 'utf-8'
_fallback_encodings = ['utf-8', 'cp1252']


def _try_decode(value):

    try:
        return str_cls(value, _encoding)

    # If the "correct" encoding did not work, try some defaults, and then just
    # obliterate characters that we can't seen to decode properly
    except (UnicodeDecodeError):
        for encoding in _fallback_encodings:
            try:
                return str_cls(value, encoding, errors='strict')
            except (UnicodeDecodeError):
                pass

    return str_cls(value, errors='replace')


def _extract_error():
    """
    Extracts the last OS error message into a python unicode string

    :return:
        A unicode string error message
    """

    error_num = errno()

    try:
        error_string = os.strerror(error_num)
    except (ValueError):
        return str_cls(error_num)

    if isinstance(error_string, str_cls):
        return error_string

    return _try_decode(error_string)


def pbkdf2(hash_algorithm, password, salt, iterations, key_length):
    """
    PBKDF2 from PKCS#5

    :param hash_algorithm:
        The string name of the hash algorithm to use: "sha1", "sha224", "sha256", "sha384", "sha512"

    :param password:
        A byte string of the password to use an input to the KDF

    :param salt:
        A cryptographic random byte string

    :param iterations:
        The numbers of iterations to use when deriving the key

    :param key_length:
        The length of the desired key in bytes

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        The derived key as a byte string
    """

    if not isinstance(password, byte_cls):
        raise TypeError(pretty_message(
            '''
            password must be a byte string, not %s
            ''',
            type_name(password)
        ))

    if not isinstance(salt, byte_cls):
        raise TypeError(pretty_message(
            '''
            salt must be a byte string, not %s
            ''',
            type_name(salt)
        ))

    if not isinstance(iterations, int_types):
        raise TypeError(pretty_message(
            '''
            iterations must be an integer, not %s
            ''',
            type_name(iterations)
        ))

    if iterations < 1:
        raise ValueError('iterations must be greater than 0')

    if not isinstance(key_length, int_types):
        raise TypeError(pretty_message(
            '''
            key_length must be an integer, not %s
            ''',
            type_name(key_length)
        ))

    if key_length < 1:
        raise ValueError('key_length must be greater than 0')

    if hash_algorithm not in set(['sha1', 'sha224', 'sha256', 'sha384', 'sha512']):
        raise ValueError(pretty_message(
            '''
            hash_algorithm must be one of "sha1", "sha224", "sha256", "sha384",
            "sha512", not %s
            ''',
            repr(hash_algorithm)
        ))

    algo = {
        'sha1': CommonCryptoConst.kCCPRFHmacAlgSHA1,
        'sha224': CommonCryptoConst.kCCPRFHmacAlgSHA224,
        'sha256': CommonCryptoConst.kCCPRFHmacAlgSHA256,
        'sha384': CommonCryptoConst.kCCPRFHmacAlgSHA384,
        'sha512': CommonCryptoConst.kCCPRFHmacAlgSHA512
    }[hash_algorithm]

    output_buffer = buffer_from_bytes(key_length)
    result = CommonCrypto.CCKeyDerivationPBKDF(
        CommonCryptoConst.kCCPBKDF2,
        password,
        len(password),
        salt,
        len(salt),
        algo,
        iterations,
        output_buffer,
        key_length
    )
    if result != 0:
        raise OSError(_extract_error())

    return bytes_from_buffer(output_buffer)


pbkdf2.pure_python = False


def rand_bytes(length):
    """
    Returns a number of random bytes suitable for cryptographic purposes

    :param length:
        The desired number of bytes

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        A byte string
    """

    if not isinstance(length, int_types):
        raise TypeError(pretty_message(
            '''
            length must be an integer, not %s
            ''',
            type_name(length)
        ))

    if length < 1:
        raise ValueError('length must be greater than 0')

    if length > 1024:
        raise ValueError('length must not be greater than 1024')

    buffer = buffer_from_bytes(length)
    result = Security.SecRandomCopyBytes(Security.kSecRandomDefault, length, buffer)
    if result != 0:
        raise OSError(_extract_error())

    return bytes_from_buffer(buffer)


# If in a future version of OS X they remove OpenSSL, this try/except block
# will fall back to the pure Python implementation, which is just slower
try:
    from .._openssl._libcrypto import libcrypto

    def _extract_openssl_error():
        """
        Extracts the last OpenSSL error message into a python unicode string

        :return:
            A unicode string error message
        """

        error_num = libcrypto.ERR_get_error()
        buffer = buffer_from_bytes(120)
        libcrypto.ERR_error_string(error_num, buffer)

        # Since we are dealing with a string, it is NULL terminated
        error_string = byte_string_from_buffer(buffer)

        return _try_decode(error_string)

    def pkcs12_kdf(hash_algorithm, password, salt, iterations, key_length, id_):
        """
        KDF from RFC7292 appendix B.2 - https://tools.ietf.org/html/rfc7292#page-19

        :param hash_algorithm:
            The string name of the hash algorithm to use: "md5", "sha1", "sha224", "sha256", "sha384", "sha512"

        :param password:
            A byte string of the password to use an input to the KDF

        :param salt:
            A cryptographic random byte string

        :param iterations:
            The numbers of iterations to use when deriving the key

        :param key_length:
            The length of the desired key in bytes

        :param id_:
            The ID of the usage - 1 for key, 2 for iv, 3 for mac

        :raises:
            ValueError - when any of the parameters contain an invalid value
            TypeError - when any of the parameters are of the wrong type
            OSError - when an error is returned by the OS crypto library

        :return:
            The derived key as a byte string
        """

        if not isinstance(password, byte_cls):
            raise TypeError(pretty_message(
                '''
                password must be a byte string, not %s
                ''',
                type_name(password)
            ))

        if not isinstance(salt, byte_cls):
            raise TypeError(pretty_message(
                '''
                salt must be a byte string, not %s
                ''',
                type_name(salt)
            ))

        if not isinstance(iterations, int_types):
            raise TypeError(pretty_message(
                '''
                iterations must be an integer, not %s
                ''',
                type_name(iterations)
            ))

        if iterations < 1:
            raise ValueError(pretty_message(
                '''
                iterations must be greater than 0 - is %s
                ''',
                repr(iterations)
            ))

        if not isinstance(key_length, int_types):
            raise TypeError(pretty_message(
                '''
                key_length must be an integer, not %s
                ''',
                type_name(key_length)
            ))

        if key_length < 1:
            raise ValueError(pretty_message(
                '''
                key_length must be greater than 0 - is %s
                ''',
                repr(key_length)
            ))

        if hash_algorithm not in set(['md5', 'sha1', 'sha224', 'sha256', 'sha384', 'sha512']):
            raise ValueError(pretty_message(
                '''
                hash_algorithm must be one of "md5", "sha1", "sha224", "sha256",
                "sha384", "sha512", not %s
                ''',
                repr(hash_algorithm)
            ))

        if id_ not in set([1, 2, 3]):
            raise ValueError(pretty_message(
                '''
                id_ must be one of 1, 2, 3, not %s
                ''',
                repr(id_)
            ))

        utf16_password = password.decode('utf-8').encode('utf-16be') + b'\x00\x00'

        digest_type = {
            'md5': libcrypto.EVP_md5,
            'sha1': libcrypto.EVP_sha1,
            'sha224': libcrypto.EVP_sha224,
            'sha256': libcrypto.EVP_sha256,
            'sha384': libcrypto.EVP_sha384,
            'sha512': libcrypto.EVP_sha512,
        }[hash_algorithm]()

        output_buffer = buffer_from_bytes(key_length)
        result = libcrypto.PKCS12_key_gen_uni(
            utf16_password,
            len(utf16_password),
            salt,
            len(salt),
            id_,
            iterations,
            key_length,
            output_buffer,
            digest_type
        )
        if result != 1:
            raise OSError(_extract_openssl_error())

        return bytes_from_buffer(output_buffer)

except (LibraryNotFoundError):

    from .._pkcs12 import pkcs12_kdf



# coding: utf-8

"""
Exceptions and compatibility shims for consistently using ctypes and cffi
"""

from __future__ import unicode_literals, division, absolute_import, print_function

import sys
import platform

from ctypes.util import find_library

from . import ffi
from ._types import str_cls, byte_cls, int_types, bytes_to_list


__all__ = [
    'array_from_pointer',
    'array_set',
    'buffer_from_bytes',
    'buffer_from_unicode',
    'buffer_pointer',
    'byte_array',
    'byte_string_from_buffer',
    'bytes_from_buffer',
    'callback',
    'cast',
    'deref',
    'errno',
    'FFIEngineError',
    'get_library',
    'is_null',
    'native',
    'new',
    'null',
    'pointer_set',
    'ref',
    'register_ffi',
    'sizeof',
    'struct',
    'struct_bytes',
    'struct_from_buffer',
    'unwrap',
    'write_to_buffer',
]


if ffi() == 'cffi':
    from cffi import FFI

    _ffi_registry = {}

    ffi = FFI()

    def register_ffi(library, ffi_obj):
        _ffi_registry[library] = ffi_obj

    def _get_ffi(library):
        if library in _ffi_registry:
            return _ffi_registry[library]
        return ffi

    def buffer_from_bytes(initializer):
        if sys.platform == 'win32':
            return ffi.new('unsigned char[]', initializer)
        return ffi.new('char[]', initializer)

    def buffer_from_unicode(initializer):
        return ffi.new('wchar_t []', initializer)

    def write_to_buffer(buffer, data, offset=0):
        buffer[offset:offset + len(data)] = data

    def buffer_pointer(buffer):
        return ffi.new('char *[]', [buffer])

    def cast(library, type_, value):
        ffi_obj = _get_ffi(library)
        return ffi_obj.cast(type_, value)

    def sizeof(library, value):
        ffi_obj = _get_ffi(library)
        return ffi_obj.sizeof(value)

    def bytes_from_buffer(buffer, maxlen=None):
        if maxlen is not None:
            return ffi.buffer(buffer, maxlen)[:]
        return ffi.buffer(buffer)[:]

    def byte_string_from_buffer(buffer):
        return ffi.string(buffer)

    def byte_array(byte_string):
        return byte_string

    def pointer_set(pointer_, value):
        pointer_[0] = value

    def array_set(array, value):
        for index, val in enumerate(value):
            array[index] = val

    def null():
        return ffi.NULL

    def is_null(point):
        if point is None:
            return True
        if point == ffi.NULL:
            return True
        if ffi.getctype(ffi.typeof(point)) == 'void *':
            return False
        if point[0] == ffi.NULL:
            return True
        return False

    def errno():
        return ffi.errno

    def new(library, type_, value=None):
        ffi_obj = _get_ffi(library)

        params = []
        if value is not None:
            params.append(value)
        if type_ in set(['BCRYPT_KEY_HANDLE', 'BCRYPT_ALG_HANDLE']):
            return ffi_obj.cast(type_, 0)
        return ffi_obj.new(type_, *params)

    def ref(value, offset=0):
        return value + offset

    def native(type_, value):
        if type_ == str_cls:
            return ffi.string(value)
        if type_ == byte_cls:
            return ffi.buffer(value)[:]
        return type_(value)

    def deref(point):
        return point[0]

    def unwrap(point):
        return point[0]

    def struct(library, name):
        ffi_obj = _get_ffi(library)
        return ffi_obj.new('%s *' % name)

    def struct_bytes(struct_):
        return ffi.buffer(struct_)[:]

    def struct_from_buffer(library, name, buffer):
        ffi_obj = _get_ffi(library)
        new_struct_pointer = ffi_obj.new('%s *' % name)
        new_struct = new_struct_pointer[0]
        struct_size = sizeof(library, new_struct)
        struct_buffer = ffi_obj.buffer(new_struct_pointer)
        struct_buffer[:] = ffi_obj.buffer(buffer, struct_size)[:]
        return new_struct_pointer

    def array_from_pointer(library, name, point, size):
        ffi_obj = _get_ffi(library)
        array = ffi_obj.cast('%s[%s]' % (name, size), point)
        total_bytes = ffi_obj.sizeof(array)
        if total_bytes == 0:
            return []
        output = []

        string_types = {
            'LPSTR': True,
            'LPCSTR': True,
            'LPWSTR': True,
            'LPCWSTR': True,
            'char *': True,
            'wchar_t *': True,
        }
        string_type = name in string_types

        for i in range(0, size):
            value = array[i]
            if string_type:
                value = ffi_obj.string(value)
            output.append(value)
        return output

    def callback(library, signature_name, func):
        ffi_obj = _get_ffi(library)
        return ffi_obj.callback(signature_name, func)

    engine = 'cffi'

else:

    import ctypes
    from ctypes import pointer, c_int, c_char_p, c_uint, c_void_p, c_wchar_p

    _pointer_int_types = int_types + (c_char_p, ctypes.POINTER(ctypes.c_byte))

    _pointer_types = {
        'void *': True,
        'wchar_t *': True,
        'char *': True,
        'char **': True,
    }
    _type_map = {
        'void *': c_void_p,
        'wchar_t *': c_wchar_p,
        'char *': c_char_p,
        'char **': ctypes.POINTER(c_char_p),
        'int': c_int,
        'unsigned int': c_uint,
        'size_t': ctypes.c_size_t,
        'uint32_t': ctypes.c_uint32,
    }
    if sys.platform == 'win32':
        from ctypes import wintypes
        _pointer_types.update({
            'LPSTR': True,
            'LPWSTR': True,
            'LPCSTR': True,
            'LPCWSTR': True,
        })
        _type_map.update({
            'BYTE': ctypes.c_byte,
            'LPSTR': c_char_p,
            'LPWSTR': c_wchar_p,
            'LPCSTR': c_char_p,
            'LPCWSTR': c_wchar_p,
            'ULONG': wintypes.ULONG,
            'DWORD': wintypes.DWORD,
            'char *': ctypes.POINTER(ctypes.c_byte),
            'char **': ctypes.POINTER(ctypes.POINTER(ctypes.c_byte)),
        })

    def _type_info(library, type_):
        is_double_pointer = type_[-3:] == ' **'
        if is_double_pointer:
            type_ = type_[:-1]
        is_pointer = type_[-2:] == ' *' and type_ not in _pointer_types
        if is_pointer:
            type_ = type_[:-2]

        is_array = type_.find('[') != -1
        if is_array:
            is_array = type_[type_.find('[') + 1:type_.find(']')]
            if is_array == '':
                is_array = True
            else:
                is_array = int(is_array)
            type_ = type_[0:type_.find('[')]

        if type_ in _type_map:
            type_ = _type_map[type_]
        else:
            type_ = getattr(library, type_)

        if is_double_pointer:
            type_ = ctypes.POINTER(type_)

        return (is_pointer, is_array, type_)

    def register_ffi(library, ffi_obj):
        pass

    def buffer_from_bytes(initializer):
        return ctypes.create_string_buffer(initializer)

    def buffer_from_unicode(initializer):
        return ctypes.create_unicode_buffer(initializer)

    def write_to_buffer(buffer, data, offset=0):
        if isinstance(buffer, ctypes.POINTER(ctypes.c_byte)):
            ctypes.memmove(buffer, data, len(data))
            return

        if offset == 0:
            buffer.value = data
        else:
            buffer.value = buffer.raw[0:offset] + data

    def buffer_pointer(buffer):
        return pointer(ctypes.cast(buffer, c_char_p))

    def cast(library, type_, value):
        is_pointer, is_array, type_ = _type_info(library, type_)

        if is_pointer:
            type_ = ctypes.POINTER(type_)
        elif is_array:
            type_ = type_ * is_array

        return ctypes.cast(value, type_)

    def sizeof(library, value):
        return ctypes.sizeof(value)

    def bytes_from_buffer(buffer, maxlen=None):
        if isinstance(buffer, _pointer_int_types):
            return ctypes.string_at(buffer, maxlen)
        if maxlen is not None:
            return buffer.raw[0:maxlen]
        return buffer.raw

    def byte_string_from_buffer(buffer):
        return buffer.value

    def byte_array(byte_string):
        return (ctypes.c_byte * len(byte_string))(*bytes_to_list(byte_string))

    def pointer_set(pointer_, value):
        pointer_.contents.value = value

    def array_set(array, value):
        for index, val in enumerate(value):
            array[index] = val

    def null():
        return None

    def is_null(point):
        return not bool(point)

    def errno():
        return ctypes.get_errno()

    def new(library, type_, value=None):
        is_pointer, is_array, type_ = _type_info(library, type_)
        if is_array:
            if is_array is True:
                type_ = type_ * value
                value = None
            else:
                type_ = type_ * is_array

        params = []
        if value is not None:
            params.append(value)
        output = type_(*params)

        if is_pointer:
            output = pointer(output)

        return output

    def ref(value, offset=0):
        if offset == 0:
            return ctypes.byref(value)
        return ctypes.cast(ctypes.addressof(value) + offset, ctypes.POINTER(ctypes.c_byte))

    def native(type_, value):
        if isinstance(value, type_):
            return value
        if sys.version_info < (3,) and type_ == int and isinstance(value, int_types):
            return value
        if isinstance(value, ctypes.Array) and value._type_ == ctypes.c_byte:
            return ctypes.string_at(ctypes.addressof(value), value._length_)
        return type_(value.value)

    def deref(point):
        return point[0]

    def unwrap(point):
        return point.contents

    def struct(library, name):
        return pointer(getattr(library, name)())

    def struct_bytes(struct_):
        return ctypes.string_at(struct_, ctypes.sizeof(struct_.contents))

    def struct_from_buffer(library, type_, buffer):
        class_ = getattr(library, type_)
        value = class_()
        ctypes.memmove(ctypes.addressof(value), buffer, ctypes.sizeof(class_))
        return ctypes.pointer(value)

    def array_from_pointer(library, type_, point, size):
        _, _, type_ = _type_info(library, type_)
        array = ctypes.cast(point, ctypes.POINTER(type_))
        output = []
        for i in range(0, size):
            output.append(array[i])
        return output

    def callback(library, signature_type, func):
        return getattr(library, signature_type)(func)

    engine = 'ctypes'


def get_library(name, dylib_name, version):
    """
    Retrieve the C library path with special handling for Mac

    :param name:
        A unicode string of the library to search the system for

    :param dylib_name:
        Mac only - a unicode string of the unversioned dylib name

    :param version:
        Mac only - a unicode string of the dylib version to use. Used on macOS
        10.15+ when the unversioned dylib is found, since unversioned
        OpenSSL/LibreSSL are just placeholders, and a versioned dylib must be
        imported. Used on macOS 10.16+ when find_library() doesn't return a
        result, due to system dylibs not being present on the filesystem any
        longer.

    :return:
        A unicode string of the path to the library
    """

    library = find_library(name)

    if sys.platform == 'darwin':
        unversioned = '/usr/lib/%s' % dylib_name
        versioned = unversioned.replace('.dylib', '.%s.dylib' % version)
        mac_ver = tuple(map(int, platform.mac_ver()[0].split('.')))
        if not library and mac_ver >= (10, 16):
            # On macOS 10.16+, find_library doesn't work, so we set a static path
            library = versioned
        elif mac_ver >= (10, 15) and library == unversioned:
            # On macOS 10.15+, we want to strongly version since unversioned libcrypto has a non-stable ABI
            library = versioned

    return library


class FFIEngineError(Exception):

    """
    An exception when trying to instantiate ctypes or cffi
    """

    pass



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

from . import backend


_backend = backend()


if _backend == 'mac':
    from ._mac.symmetric import (
        aes_cbc_no_padding_decrypt,
        aes_cbc_no_padding_encrypt,
        aes_cbc_pkcs7_decrypt,
        aes_cbc_pkcs7_encrypt,
        des_cbc_pkcs5_decrypt,
        des_cbc_pkcs5_encrypt,
        rc2_cbc_pkcs5_decrypt,
        rc2_cbc_pkcs5_encrypt,
        rc4_decrypt,
        rc4_encrypt,
        tripledes_cbc_pkcs5_decrypt,
        tripledes_cbc_pkcs5_encrypt,
    )

elif _backend == 'win' or _backend == 'winlegacy':
    from ._win.symmetric import (
        aes_cbc_no_padding_decrypt,
        aes_cbc_no_padding_encrypt,
        aes_cbc_pkcs7_decrypt,
        aes_cbc_pkcs7_encrypt,
        des_cbc_pkcs5_decrypt,
        des_cbc_pkcs5_encrypt,
        rc2_cbc_pkcs5_decrypt,
        rc2_cbc_pkcs5_encrypt,
        rc4_decrypt,
        rc4_encrypt,
        tripledes_cbc_pkcs5_decrypt,
        tripledes_cbc_pkcs5_encrypt,
    )

else:
    from ._openssl.symmetric import (
        aes_cbc_no_padding_decrypt,
        aes_cbc_no_padding_encrypt,
        aes_cbc_pkcs7_decrypt,
        aes_cbc_pkcs7_encrypt,
        des_cbc_pkcs5_decrypt,
        des_cbc_pkcs5_encrypt,
        rc2_cbc_pkcs5_decrypt,
        rc2_cbc_pkcs5_encrypt,
        rc4_decrypt,
        rc4_encrypt,
        tripledes_cbc_pkcs5_decrypt,
        tripledes_cbc_pkcs5_encrypt,
    )


__all__ = [
    'aes_cbc_no_padding_decrypt',
    'aes_cbc_no_padding_encrypt',
    'aes_cbc_pkcs7_decrypt',
    'aes_cbc_pkcs7_encrypt',
    'des_cbc_pkcs5_decrypt',
    'des_cbc_pkcs5_encrypt',
    'rc2_cbc_pkcs5_decrypt',
    'rc2_cbc_pkcs5_encrypt',
    'rc4_decrypt',
    'rc4_encrypt',
    'tripledes_cbc_pkcs5_decrypt',
    'tripledes_cbc_pkcs5_encrypt',
]
